
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "CBP.h"

#define M_PI2	(M_PI*M_PI)
#define delta2	(delta*delta)
#define k2	((double)k*(double)k)

double	Ramachandran(delta,k)
double	delta;
int	k;
{
	if (k==0)
	    return 1.0/(4.0*delta2);
	else if (k&1)
	    return -1.0/(M_PI2*delta2*k2);
	else
	    return 0.0;
}

double	Shepp(delta,k)
double	delta;
int	k;
{
	return 2.0/(M_PI2*delta2*(1.0-4.0*k2));
}

double	Chesler(delta,k)
double	delta;
int	k;
{
	if (k==0)
	    return (1.0/4.0-1.0/M_PI2)/(2.0*delta2);
	else if (k==1 || k==(-1))
	    return (1.0/8.0-1.0/M_PI2)/(2.0*delta2);
	else if (k&1)
	    return -1.0/(2.0*M_PI2*delta2*k2);
	else
	    return -(k2+1.0)/(2.0*M_PI2*delta2*(k2-1.0)*(k2-1.0));
}

#ifndef	SIMPLE_CONVOLUTION
static void	FFT(L,Z,E,sign)
int		L;
Complex		*Z,*E;
double		sign;
{
	int	j,k,m,n,p,q,L1=L-1,L2=L>>1;
	Complex	z,*e,*x,*y;
	double	Cos,Sin,re,im;

	for (j=1; j<L1; j++) {
	    k=0; for (m=L2, n=1; n<=j; n<<=1, m>>=1) if (j&n) k|=m;
	    if (j<k) {
		z=Z[j]; Z[j]=Z[k]; Z[k]=z;
	    }
	}
	for (j=2, k=1, m=L2, n=L; n>1; n=m, m>>=1, k=j, j<<=1)
	    for (e=E, p=0; p<k; p++, e+=n) {
		Cos=e->re; Sin=e->im*sign;
		for (x=Z+p, q=0; q<m; q++, x+=j) {
		    y=x+k; re=y->re*Cos-y->im*Sin;
			   im=y->re*Sin+y->im*Cos; y->re=x->re-re; x->re+=re;
						   y->im=x->im-im; x->im+=im;
		}
	    }
}
#endif

#define ALLOC(noe,ptr)	malloc((noe)*sizeof(*(ptr)))

CBP	*SetUpCBP(N,M,delta,r0,theta0)
int	N,M;
double	delta,r0,theta0;
{
	int	m,n;
	CBP	*cbp;
	double	theta,R,Filter();

#ifndef	SIMPLE_CONVOLUTION
	int	L,L2;
	double	phase;

	for (L=2; L<N; L<<=1) ; L2=L<<1;
#else
	int	N21=N*2-1,N1=N-1;
#endif

	if ((cbp     =(CBP     *)ALLOC(1  ,cbp     ))==NULL ||
	    (cbp->pq =(Float  **)ALLOC(M  ,cbp->pq ))==NULL ||
#ifndef	SIMPLE_CONVOLUTION
	    (cbp->G  =(Complex *)ALLOC(L2 ,cbp->G  ))==NULL ||
	    (cbp->Z  =(Complex *)ALLOC(L2 ,cbp->Z  ))==NULL ||
	    (cbp->E  =(Complex *)ALLOC(L2 ,cbp->E  ))==NULL ||
#else
	    (cbp->g  =(double  *)ALLOC(N21,cbp->g  ))==NULL ||
	    (cbp->z  =(double  *)ALLOC(N,  cbp->z  ))==NULL ||
#endif
	    (cbp->Cos=(double  *)ALLOC(M  ,cbp->Cos))==NULL ||
	    (cbp->Sin=(double  *)ALLOC(M  ,cbp->Sin))==NULL) return NULL;

#ifdef	CONCENTRIC_SYMMETRY
	if ((cbp->pq[0]=(Float *)ALLOC(N+1,cbp->pq[0]))==NULL) return NULL;

	for (m=1; m<M; m++) cbp->pq[m]=cbp->pq[0];

	for (n=0; n<=N; n++) cbp->pq[0][n]=0.0;
#else
	for (m=0; m<M; m++) {
	    if ((cbp->pq[m]=(Float *)ALLOC(N+1,cbp->pq[m]))==NULL) return NULL;

	    for (n=0; n<=N; n++) cbp->pq[m][n]=0.0;
	}
#endif
	cbp->N=N;
#ifndef	SIMPLE_CONVOLUTION
	cbp->L=L;
#endif
	cbp->delta=delta;
	cbp->r0delta=r0/delta;
	R=r0+delta*(double)(N-1); if (R>(-r0)) R=(-r0); cbp->R2=R*R;

#ifndef	SIMPLE_CONVOLUTION
	for (n=0; n<L2; n++) {
	    cbp->G[n].re=Filter(delta,n-L);
	    cbp->G[n].im=0.0;

	    phase=M_PI*(double)n/(double)L2; cbp->E[n].re=cos(phase);
					     cbp->E[n].im=sin(phase);
	}
	FFT(L2,cbp->G,cbp->E,-1.0);
#else
	for (n=(-N1); n<=N1; n++) cbp->g[n+N1]=Filter(delta,n);
#endif
	cbp->M=M;
	cbp->theta0=theta0;
	cbp->M_PIM=M_PI/(double)M;

	for (m=0; m<M; m++) {
	    theta=theta0+cbp->M_PIM*(double)m; cbp->Cos[m]=cos(theta);
					       cbp->Sin[m]=sin(theta);
	}
	return cbp;
}

void	Convolution(cbp)
CBP	*cbp;
{
	int	M,m,n,N=cbp->N;
	Float	**pq,*p,*q;

#ifndef	SIMPLE_CONVOLUTION
	int	L=cbp->L,L2=L<<1;
	Complex	*Z,*G,c,*cbpZ=cbp->Z,*cbpG=cbp->G,*cbpE=cbp->E;
	double	deltaL2=cbp->delta/(double)L2;
#else
	int	j;
	double	*zj,*gj,z,*g,
		*cbpz=cbp->z,*cbpg=(&(cbp->g[N-1])),delta=cbp->delta;
#endif

#ifdef	CONCENTRIC_SYMMETRY
	M=1;
#else
	M=cbp->M;
#endif

	for (pq=cbp->pq, m=0; m<M; m++, pq++) {
#ifndef	SIMPLE_CONVOLUTION
	    for (Z=cbpZ,  n=0; n<L; n++,      Z++)   Z->re=      Z->im=0.0;
	    for (p=(*pq), n=0; n<N; n++, p++, Z++) { Z->re=(*p); Z->im=0.0; }
	    for (            ; n<L; n++,      Z++)   Z->re=      Z->im=0.0;

	    FFT(L2,cbpZ,cbpE,-1.0);

	    for (Z=cbpZ, G=cbpG, n=0; n<L2; n++, G++, Z++) {
		c.re=G->re*Z->re-G->im*Z->im;
		c.im=G->re*Z->im+G->im*Z->re; *Z=c;
	    }
	    FFT(L2,cbpZ,cbpE, 1.0);

	    for (q=(*pq), Z=cbpZ, n=0; n<N; n++, Z++, q++) *q=deltaL2*Z->re;
#else
	    for (zj=cbpz, gj=cbpg, j=0; j<N; j++, gj++, zj++) {
		z=0.0;
		for (p=(*pq), g=gj, n=0; n<N; n++, g--, p++) z+=(*g*(*p));
		*zj=delta*z;
	    }
	    for (q=(*pq), zj=cbpz, j=0; j<N; j++, zj++, q++) *q=(*zj);
#endif
	}
}

double	BackProjection(cbp,x,y)
CBP	*cbp;
double	x,y;
{
	double	r0,*Cos,*Sin,r,f=0.0;
	int	M,m,n;
	Float	**pq,*q;

	if (x*x+y*y<=cbp->R2) {
	    x/=cbp->delta;
	    y/=cbp->delta; r0=cbp->r0delta;
	    M=cbp->M; pq=cbp->pq; Cos=cbp->Cos; Sin=cbp->Sin;
	    for (m=0; m<M; m++, Sin++, Cos++, pq++) {
		r=x*(*Cos)+y*(*Sin)-r0; n=(int)r; q=(*pq);
		f+=((r-(double)n)*(q[n+1]-q[n])+q[n]);
	    }
	}
	return f*cbp->M_PIM;
}

Float	**BPImage(cbp)
CBP	*cbp;
{
	Float	**f,**fv,*fh,*q;
	int	v,h,*h0,*h0v,*Oh,*Ohv,m,n,N=cbp->N,M=cbp->M;
	double	x,Cos,Sin,rv,rh,N12=(double)(N-1)/2.0,
				R2=cbp->R2/(cbp->delta*cbp->delta),
				r0=cbp->r0delta,
				M_PIM=cbp->M_PIM;
#ifdef	__INTEL_COMPILER
	double	y;
	int	v1=N-1,v2=0;

	if ((f=fv=(Float **)ALLOC(N,f))==NULL ||
	    (h0=h0v=(int *)ALLOC(N,h0))==NULL ||
	    (Oh=Ohv=(int *)ALLOC(N,Oh))==NULL) return NULL;

	for (v=0; v<N; v++, fv++, h0v++, Ohv++) {
	    if ((*fv=fh=(Float *)ALLOC(N,*f))==NULL) return NULL;

	    for (h=N; h>0; h--, fh++) *fh=0.0;

	    y=N12-(double)v;
	    if ((x=R2-y*y)>=0.0) {
		x=sqrt(x); *Ohv=(int)(N12+x)-(*h0v=(int)ceil(N12-x));

		if (v<v1) v1=v;
		if (v>v2) v2=v;
	    }
	}
	h0+=v1; Oh+=v1;
#else
	double	y=sqrt(R2);
	int	v1=(int)ceil(N12-y),v2=(int)(N12+y);

	if ((f=fv=(Float **)ALLOC(N,f))==NULL) return NULL;

	for (v=0; v<N; v++, fv++) {
	    if ((*fv=fh=(Float *)ALLOC(N,*f))==NULL) return NULL;

	    for (h=N; h>0; h--, fh++) *fh=0.0;
	}
	v=v2-v1+1;
	if ((h0=h0v=(int *)ALLOC(v,h0))==NULL ||
	    (Oh=Ohv=(int *)ALLOC(v,Oh))==NULL) return NULL;

	for (v=v1; v<=v2; v++, h0v++, Ohv++) {
	    y=N12-(double)v; x=sqrt(R2-y*y);
	    *Ohv=(int)(N12+x)-(*h0v=(int)ceil(N12-x));
	}
#endif
	for (m=0; m<M; m++) {
	    q=cbp->pq[m]; Cos=cbp->Cos[m]; Sin=cbp->Sin[m];

	    rv=(-r0)+(N12-(double)v1)*Sin; fv=f+v1; h0v=h0; Ohv=Oh;
	    for (v=v1; v<=v2; v++, rv-=Sin, fv++, h0v++, Ohv++) {
		h=(*h0v); rh=rv+((double)h-N12)*Cos; fh=(*fv)+h;
		for (h=(*Ohv); h>=0; h--, rh+=Cos, fh++) {
		    n=(int)rh; *fh+=((rh-(double)n)*(q[n+1]-q[n])+q[n]);
		}
	    }
	}
	for (fv=f+(v=v1), h0v=h0, Ohv=Oh; v<=v2; v++, fv++, h0v++, Ohv++)
	    for (fh=(*fv)+(*h0v), h=(*Ohv); h>=0; h--, fh++) *fh*=M_PIM;

#ifdef	__INTEL_COMPILER
	free(Oh-v1); free(h0-v1);
#else
	free(Oh); free(h0);
#endif
	return f;
}

int	ReGridingCP(cbp)
CBP	*cbp;
{
	int	N,M,segments,m,n,n0,s;
	Float	**pq,**new1,*old1,*old2,*new2;
	double	r0,ratio,theta,NM_PI2=(double)cbp->N*M_PI/2.0;

	if ((double)cbp->M>=NM_PI2) return 1;

	N=cbp->N;
	M=cbp->M; segments=(int)ceil(NM_PI2/(double)M); cbp->M=M*segments;

	if ((pq   =(Float **)ALLOC(M+1,pq   ))==NULL ||
	    (pq[M]=(Float  *)ALLOC(N+1,pq[M]))==NULL) return 0;

	for (m=0; m<M; m++) pq[m]=cbp->pq[m];

	for (n=0; n<=N; n++) {
	    r0=(-2.0*cbp->r0delta-(double)n);
	    if (r0>=0.0 && r0<=(double)(N-1)) {
		n0=(int)r0;
		pq[M][n]=pq[0][n0]+(r0-(double)n0)*(pq[0][n0+1]-pq[0][n0]);
	    }
	    else
		pq[M][n]=0.0;
	}

	free(cbp->pq);
	if ((cbp->pq=new1=(Float **)ALLOC(cbp->M,cbp->pq))==NULL) return 0;

	for (old1=pq[0], m=1; m<=M; m++, old1=old2) {
	    *new1=old1; old2=pq[m];
	    for (++new1, s=1; s<segments; s++, new1++) {
		if ((*new1=new2=(Float *)ALLOC(N+1,new2))==NULL) return 0;

		ratio=(double)s/(double)segments;
		for (n=0; n<=N; n++) *(new2++)=old1[n]+ratio*(old2[n]-old1[n]);
	    }
	}
	free(pq[M]); free(pq);

	free(cbp->Cos);
	free(cbp->Sin);
	if ((cbp->Cos=(double *)ALLOC(cbp->M,cbp->Cos))==NULL ||
	    (cbp->Sin=(double *)ALLOC(cbp->M,cbp->Sin))==NULL) return 0;

	cbp->M_PIM=M_PI/(double)cbp->M;

	for (m=0; m<cbp->M; m++) {
	    theta=cbp->theta0+cbp->M_PIM*(double)m; cbp->Cos[m]=cos(theta);
						    cbp->Sin[m]=sin(theta);
	}
	return 1;
}

void	ClearCBP(cbp,f)
CBP	*cbp;
Float	**f;
{
	int	n,m,N=cbp->N,M=cbp->M;
	Float	**pq=cbp->pq;

	if (f!=NULL) {
	    for (n=0; n<N; n++) free(f[n]); free(f);
	}

	free(cbp->Sin); free(cbp->Cos);

#ifndef	SIMPLE_CONVOLUTION
	free(cbp->E); free(cbp->Z); free(cbp->G);
#else
	free(cbp->z); free(cbp->g);
#endif

#ifdef	CONCENTRIC_SYMMETRY
{
	int	m0;

	for (m=0; m<M; m++) {
	    for (m0=0; m0<m; m0++) if (pq[m]==pq[m0]) break;

	    if (m==m0) free(pq[m]);
	}
}
#else
	for (m=0; m<M; m++) free(pq[m]);
#endif
	free(pq); free(cbp);
}
