
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "FBP.h"

double	Ramachandran(delta,z)
double	delta,z;
{
	z=fabs(z); return (z<1.0/(2.0*delta))?z:0.0;
}

double	Shepp(delta,z)
double	delta,z;
{
	return fabs(sin(M_PI*delta*z))/(M_PI*delta);
}

double	Chesler(delta,z)
double	delta,z;
{
	return fabs(z)*(1.0+cos(2.0*M_PI*delta*z))/2.0;
}

#define ALLOC(noe,ptr)	malloc((noe)*sizeof(*(ptr)))

FBP	*SetUpFBP(N,M,delta,r0,theta0)
int	N,M;
double	delta,r0,theta0;
{
	int	L,m,n;
	FBP	*fbp;
	double	epsilon,phase,G,theta,Filter();
	Complex	A,B;

	for (L=0; 1<<L<N; L++) ; N=1<<L;

	if ((fbp     =(FBP     *)ALLOC(1,fbp     ))==NULL ||
	    (fbp->pq =(Float  **)ALLOC(M,fbp->pq ))==NULL ||
	    (fbp->Z  =(Complex *)ALLOC(N,fbp->Z  ))==NULL ||
	    (fbp->E  =(Complex *)ALLOC(N,fbp->E  ))==NULL ||
	    (fbp->ABG=(Complex *)ALLOC(N,fbp->ABG))==NULL ||
	    (fbp->Cos=(double  *)ALLOC(M,fbp->Cos))==NULL ||
	    (fbp->Sin=(double  *)ALLOC(M,fbp->Sin))==NULL) return NULL;

#ifdef	CONCENTRIC_SYMMETRY
	if ((fbp->pq[0]=(Float *)ALLOC(N,fbp->pq[0]))==NULL) return NULL;

	for (m=1; m<M; m++) fbp->pq[m]=fbp->pq[0];

	for (n=0; n<N; n++) fbp->pq[0][n]=0.0;
#else
	for (m=0; m<M; m++) {
	    if ((fbp->pq[m]=(Float *)ALLOC(N,fbp->pq[m]))==NULL) return NULL;

	    for (n=0; n<N; n++) fbp->pq[m][n]=0.0;
	}
#endif
	fbp->N=N;
	fbp->delta=delta;

	epsilon=(r0/delta+(double)(N-1)/2.0)/(double)N;
	phase=M_PI*epsilon*(double)(N-1); A.re=cos(phase)/(double)N;
					  A.im=sin(phase)/(double)N;
	for (n=0; n<N; n++) {
	    phase=M_PI*(double)n/(double)N; fbp->E[n].re=cos(phase);
					    fbp->E[n].im=sin(phase);
	    phase=2.0*M_PI*epsilon*(double)n; B.re=  cos(phase) ;
					      B.im=(-sin(phase));

	    G=Filter(delta,((double)(1-N)/2.0+(double)n)/(delta*(double)N));

	    fbp->ABG[n].re=(A.re*B.re-A.im*B.im)*G;
	    fbp->ABG[n].im=(A.re*B.im+A.im*B.re)*G;
	}
	fbp->M=M;
	fbp->theta0=theta0;
	fbp->M_PIM=M_PI/(double)M;

	for (m=0; m<M; m++) {
	    theta=theta0+fbp->M_PIM*(double)m; fbp->Cos[m]=cos(theta);
					       fbp->Sin[m]=sin(theta);
	}
	return fbp;
}

static void	FFT(L,Z,E,sign)
int		L;
Complex		*Z,*E;
double		sign;
{
	int	j,k,m,n,p,q,L1=L-1,L2=L>>1;
	Complex	z,*e,*x,*y;
	double	Cos,Sin,re,im;

	for (j=1; j<L1; j++) {
	    k=0; for (m=L2, n=1; n<=j; n<<=1, m>>=1) if (j&n) k|=m;
	    if (j<k) {
		z=Z[j]; Z[j]=Z[k]; Z[k]=z;
	    }
	}
	for (j=2, k=1, m=L2, n=L; n>1; n=m, m>>=1, k=j, j<<=1)
	    for (e=E, p=0; p<k; p++, e+=n) {
		Cos=e->re; Sin=e->im*sign;
		for (x=Z+p, q=0; q<m; q++, x+=j) {
		    y=x+k; re=y->re*Cos-y->im*Sin;
			   im=y->re*Sin+y->im*Cos; y->re=x->re-re; x->re+=re;
						   y->im=x->im-im; x->im+=im;
		}
	    }
}

void	Filtering(fbp)
FBP	*fbp;
{
	int	M,m,n,N=fbp->N;
	Float	**pq,*p,*q;
	Complex	*Z,*E,*ABG,c,*fbpZ=fbp->Z,*fbpE=fbp->E,*fbpABG=fbp->ABG;

#ifdef	CONCENTRIC_SYMMETRY
	M=1;
#else
	M=fbp->M;
#endif
	for (pq=fbp->pq, m=0; m<M; m++, pq++) {
	    for (p=(*pq), Z=fbpZ, E=fbpE, n=0; n<N; n++, E++, Z++, p++)
		if (n&1) { Z->re=(-E->re*(*p));
			   Z->im=  E->im*(*p) ; } else
			 { Z->re=  E->re*(*p) ;
			   Z->im=(-E->im*(*p)); }

	    FFT(N,fbpZ,fbpE,-1.0);

	    for (Z=fbpZ, ABG=fbpABG, n=0; n<N; n++, ABG++, Z++) {
		c.re=ABG->re*Z->re-ABG->im*Z->im;
		c.im=ABG->re*Z->im+ABG->im*Z->re; *Z=c;
	    }
	    FFT(N,fbpZ,fbpE, 1.0);

	    for (q=(*pq), Z=fbpZ, E=fbpE, n=0; n<N; n++, E++, Z++, q++)
		if (n&1) *q=(-E->re*Z->re+E->im*Z->im); else
			 *q=  E->re*Z->re-E->im*Z->im ;
	}
}

double	BackProjection(fbp,x,y)
FBP	*fbp;
double	x,y;
{
	double	*Cos,*Sin,rm,r,N2=(double)fbp->N/2.0,f=0.0;
	Float	**pq,*q;
	int	M,m,n;

	x/=fbp->delta;
	y/=fbp->delta;
	if (x*x+y*y<=N2*N2) {
	    M=fbp->M; pq=fbp->pq; Cos=fbp->Cos; Sin=fbp->Sin;
	    for (m=0; m<M; m++, Sin++, Cos++, pq++) {
		rm=x*(*Cos)+y*(*Sin);
		if (rm>(-N2) && rm<N2) {
		    q=(*pq);
		    if (rm<=0.5-N2)
			f+=(2.0*(N2+rm)*q[0]);
		    else
		    if (rm>=N2-0.5)
			f+=(2.0*(N2-rm)*q[fbp->N-1]);
		    else {
			r=rm+N2-0.5; n=(int)r;
			f+=((r-(double)n)*(q[n+1]-q[n])+q[n]);
		    }
		}
	    }
	}
	return f*fbp->M_PIM;
}

Float	**BPImage(fbp)
FBP	*fbp;
{
	Float	**f,**fv,*fh,*q;
	int	*h1,*h2,v,h,h2v,m,n,N=fbp->N,v1=N-1,v2=0,M=fbp->M;
	double	*xy,Rv2,Rv,Cos,Sin,ySin,rm,r,
		N2=(double)N/2.0,N12=N2-0.5,R2=N2*N2,M_PIM=fbp->M_PIM;

	if ((f =(Float **)ALLOC(N,f ))==NULL ||
	    (xy=(double *)ALLOC(N,xy))==NULL ||
	    (h1=(int    *)ALLOC(N,h1))==NULL ||
	    (h2=(int    *)ALLOC(N,h2))==NULL) return NULL;

	for (fv=f, v=0; v<N; v++, fv++) {
	    if ((*fv=(Float *)ALLOC(N,*fv))==NULL) return NULL;

	    for (fh=(*fv), h=0; h<N; h++, fh++) *fh=0.0;

	    xy[v]=(double)v-N12;
	    if ((Rv2=R2-xy[v]*xy[v])>=0.0) {
		if (v<v1) v1=v;
		if (v>v2) v2=v; Rv=sqrt(Rv2); h1[v]=(int)ceil(N12-Rv);
					      h2[v]=(int)    (N12+Rv);
	    }
	}
	for (m=0; m<M; m++) {
	    q=fbp->pq[m]; Cos=fbp->Cos[m]; Sin=fbp->Sin[m];
	    for (v=v1, fv=f+v; v<=v2; v++, fv++) {
		ySin=xy[v]*Sin;
		for (h2v=h2[v], h=h1[v], fh=(*fv)+h; h<=h2v; h++, fh++) {
		    rm=xy[h]*Cos-ySin;
		    if (rm>(-N2) && rm<N2)
			if (rm<=(-N12))
			    *fh+=(2.0*(N2+rm)*q[0]);
			else
			if (rm>=N12)
			    *fh+=(2.0*(N2-rm)*q[N-1]);
			else {
			    r=rm+N12; n=(int)r;
			    *fh+=((r-(double)n)*(q[n+1]-q[n])+q[n]);
			}
		}
	    }
	}
	for (           v=v1   , fv=  f  +v; v<=v2 ; v++, fv++)
	for (h2v=h2[v], h=h1[v], fh=(*fv)+h; h<=h2v; h++, fh++) *fh*=M_PIM;

	free(h2); free(h1); free(xy); return f;
}

int	ReGridingFP(fbp)
FBP	*fbp;
{
	int	N,M,segments,m,n,s;
	Float	**pq,**new1,*old1,*old2,*new2;
	double	ratio,theta,NM_PI2=(double)fbp->N*M_PI/2.0;

	if ((double)fbp->M>=NM_PI2) return 1;

	N=fbp->N;
	M=fbp->M; segments=(int)ceil(NM_PI2/(double)M); fbp->M=M*segments;

	if ((pq   =(Float **)ALLOC(M+1,pq   ))==NULL ||
	    (pq[M]=(Float  *)ALLOC(N,  pq[M]))==NULL) return 0;

	for (m=0; m<M; m++) pq[m]=fbp->pq[m];

	for (n=0; n<N; n++) pq[M][n]=pq[0][N-1-n];

	free(fbp->pq);
	if ((fbp->pq=new1=(Float **)ALLOC(fbp->M,fbp->pq))==NULL) return 0;

	for (old1=pq[0], m=1; m<=M; m++, old1=old2) {
	    *new1=old1; old2=pq[m];
	    for (++new1, s=1; s<segments; s++, new1++) {
		if ((*new1=new2=(Float *)ALLOC(N,new2))==NULL) return 0;

		ratio=(double)s/(double)segments;
		for (n=0; n<N; n++) *(new2++)=old1[n]+ratio*(old2[n]-old1[n]);
	    }
	}
	free(pq[M]); free(pq);

	free(fbp->Cos);
	free(fbp->Sin);
	if ((fbp->Cos=(double *)ALLOC(fbp->M,fbp->Cos))==NULL ||
	    (fbp->Sin=(double *)ALLOC(fbp->M,fbp->Sin))==NULL) return 0;

	fbp->M_PIM=M_PI/(double)fbp->M;

	for (m=0; m<fbp->M; m++) {
	    theta=fbp->theta0+fbp->M_PIM*(double)m; fbp->Cos[m]=cos(theta);
						    fbp->Sin[m]=sin(theta);
	}
	return 1;
}

void	ClearFBP(fbp,f)
FBP	*fbp;
Float	**f;
{
	int	n,m,N=fbp->N,M=fbp->M;
	Float	**pq=fbp->pq;

	if (f!=NULL) {
	    for (n=0; n<N; n++) free(f[n]); free(f);
	}

	free(fbp->Sin);
	free(fbp->Cos); free(fbp->ABG); free(fbp->E); free(fbp->Z);

#ifdef	CONCENTRIC_SYMMETRY
{
	int	m0;

	for (m=0; m<M; m++) {
	    for (m0=0; m0<m; m0++) if (pq[m]==pq[m0]) break;

	    if (m==m0) free(pq[m]);
	}
}
#else
	for (m=0; m<M; m++) free(pq[m]);
#endif
	free(pq); free(fbp);
}
