
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "rif.h"
#include "sif.h"

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

#define DESC_LEN	256

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	cnt,idx,Nx,Ny,x,y,BPS;
	Cell	**cell;
	double	*sum,*ptr,base,step,val,min,max;
	char	desc[DESC_LEN];

	if ((cnt=(argc<=5)?0:
		 ((argc-3)%3==0)?(argc-3)/3:
		 ((argc-5)%3==0)?(argc-5)/3:0)==0)
	    Error(
"usage : atvi TIFF1 base1 step1 {TIFF2 base2 step2 ...} {base3 step3} BPS TIFF3"
	    );

	for (idx=0; idx<cnt; idx++) {
	    if (idx==0) {
		ReadImageFile(argv[1],&Nx,&Ny,NULL,&cell,NULL);

		if ((sum=(double *)malloc(Nx*Ny*sizeof(double)))==NULL)
		    Error("memory allocation error.");

		ptr=sum; for (y=0; y<Ny; y++) for (x=0; x<Nx; x++) *(ptr++)=0.0;
	    }
	    else {
		ReadImageFile(argv[1+idx*3],&x,&y,NULL,&cell,NULL);

		if (x!=Nx || y!=Ny) Error("cell size not match.");
	    }
	    base=atof(argv[2+idx*3]);
	    step=atof(argv[3+idx*3]);

	    ptr=sum;
	    if (idx<cnt-1) {
		for (y=0; y<Ny; y++) {
		    for (x=0; x<Nx; x++)
			*(ptr++)+=(base+step*(double)cell[y][x]);

		    free(cell[y]);
		}
		free(cell);
	    }
	    else
		for (y=0; y<Ny; y++) for (x=0; x<Nx; x++) {
		    val=(*(ptr++)+=(base+step*(double)cell[y][x]));

		    if (x==0 && y==0)
			min=max=val;
		    else
			if (val<min) min=val; else if (val>max) max=val;
		}
	}


	if ((BPS=atoi(argv[argc-2]))<=0 || BPS>8*sizeof(Cell))
	    Error("bad bits per sample.");

	if (argc==3+cnt*3) {
	    base=min;
	    if (min==max) {
		fputs("bad value range (warning).\n",stderr);

		step=1.0;
	    }
	    else
		step=(max-min)/(double)((1<<BPS)-1);

	    sprintf(desc,"%lf\t%lf",min,max);
	}
	else {
	    base=atof(argv[argc-4]);
	    if ((step=atof(argv[argc-3]))==0.0) Error("bad value step.");

	    sprintf(desc,"%lf\t%lf\n%lf\t%lf",min,max,base,step);

	    min=base; max=base+step*(double)((1<<BPS)-1);
	}
	puts(desc);

	ptr=sum;
	for (y=0; y<Ny; y++)
	for (x=0; x<Nx; x++) {
	    if ((val=(*ptr++))<min) val=min; else if (val>max) val=max;

	    cell[y][x]=(Cell)floor((val-base)/step+0.5);
	}
	StoreImageFile(argv[argc-1],Nx,Ny,BPS,cell,desc); return 0;
}
