
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "cell.h"
#include "rif.h"

#ifndef	Value
#define Value	float
#endif

static void	Error(msg)
char		*msg;
{
	(void)fputs(msg,stderr); (void)fputc('\n',stderr); exit(1);
}

#define ALLOC(type,size)	(type *)malloc(sizeof(type)*(size))

static int	max(a,b)
int		a,b;
{
	return (a>b)?a:b;
}

typedef struct {
		char	*path;
		double	base,step;
	} PBS;

#define PATH_LEN	256

static PBS	*CheckPBSNxyz(Nx,Ny,Nz)
int		*Nx,*Ny,*Nz;
{
	PBS	*pbs;
	int	z,nx,ny;
	char	path[PATH_LEN];

	if (scanf("%d",Nz)!=1 || *Nz<=0) Error("bad slice number.");

	if ((pbs=ALLOC(PBS,*Nz))==NULL)
	    Error("path, base and/or step memory allocation error.");

	*Nx=(*Ny)=0;
	for (z=0; z<*Nz; z++) {
	    if (scanf("%s %lf %lf",path,&(pbs[z].base),&(pbs[z].step))!=3)
		Error("path, base and/or step read error.");

	    ReadImageFile(path,&nx,&ny,NULL,NULL,NULL);
	    *Nx=max(*Nx,nx);
	    *Ny=max(*Ny,ny);

	    if ((pbs[z].path=ALLOC(char,strlen(path)+1))==NULL)
		Error("path memory allocation error.");

	    (void)strcpy(pbs[z].path,path);
	}
	return pbs;
}

static void	ScanValue(pbs,x1,y1,x2,y2,value)
PBS		*pbs;
int		x1,y1,x2,y2;
Value		**value;
{
	int	nx,ny,x,y;
	Cell	**cell;
	double	base=pbs->base,step=pbs->step;

	ReadImageFile(pbs->path,&nx,&ny,NULL,&cell,NULL);

	for (y=y1; y<=y2; y++)
	for (x=x1; x<=x2; x++)
	    value[y][x]=base+step*(double)((x<nx && y<ny)?cell[y][x]:0);

	while (--ny>=0) free(cell[ny]); free(cell);
}

static int	min(a,b)
int		a,b;
{
	return (a<b)?a:b;
}

static void	StoreRange(Nz1,Nz2,dz1,dz2,z11,z12,Z21,Z22)
int		Nz1,Nz2,dz1,dz2,*z11,*z12,**Z21,**Z22;
{
	int	nz,z1;

	*z11=max(0,-dz2); *z12=min(Nz1,Nz2-dz1)-1; nz=(*z12)-(*z11)+1;

	if ((*Z21=ALLOC(int,nz))==NULL ||
	    (*Z22=ALLOC(int,nz))==NULL) Error("range memory allocation error.");

	*Z21-=(*z11); *Z22-=(*z11);

	for (z1=(*z11); z1<=(*z12); z1++) {
	    (*Z21)[z1]=max(0,z1+dz1); (*Z22)[z1]=min(Nz2-1,z1+dz2);
	}
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	PBS	*pbs1,*pbs2;
	int	Nx1,Ny1,Nz1,Nx2,Ny2,Nz2,dx1,dy1,dz1,dx2,dy2,dz2,
		nx,ny,Dz,dz,dy,dx,
		x11,x12,*X21,*X22,y11,y12,*Y21,*Y22,z11,z12,*Z21,*Z22,
		y1,x21,x22,y21,y22,y2,z21,z22,z1,z2,Z2,Y2,x1,x2,X2,nz;
	double	***Sqrs,***Sz,***sz,**Sy,**sy,*Sx,*sx,diff;
	char	*cmae="count memory allocation error.",
		*vmae="value memory allocation error";
	Value	**Value1,***Value2,**value2,
		**v1y,*V1x,**v2y,*V2x,*v1x,value1,*v2x;

	if (argc!=7) Error("usage : ctvi3 dx1 dy1 dz1 dx2 dy2 dz2");

	pbs1=CheckPBSNxyz(&Nx1,&Ny1,&Nz1);
	pbs2=CheckPBSNxyz(&Nx2,&Ny2,&Nz2);

	if ((dx1=atoi(argv[1]))<=(-Nx1) ||
	    (dx2=atoi(argv[4]))>=  Nx2  || dx1>dx2 ||
	    (dy1=atoi(argv[2]))<=(-Ny1) ||
	    (dy2=atoi(argv[5]))>=  Ny2  || dy1>dy2 ||
	    (dz1=atoi(argv[3]))<=(-Nz1) ||
	    (dz2=atoi(argv[6]))>=  Nz2  || dz1>dz2)
	    Error("bad range of distance.");

	nx=dx2-dx1+1; ny=dy2-dy1+1; Dz=dz2-dz1+1;

	if ((Sqrs=ALLOC(double **,Dz))==NULL) Error(cmae);

	Sqrs-=dz1;

	for (dz=dz1; dz<=dz2; dz++) {
	    if ((Sqrs[dz]=ALLOC(double *,ny))==NULL) Error(cmae);

	    Sqrs[dz]-=dy1;

	    for (dy=dy1; dy<=dy2; dy++) {
		if ((Sqrs[dz][dy]=ALLOC(double,nx))==NULL) Error(cmae);

		Sqrs[dz][dy]-=dx1;

		for (dx=dx1; dx<=dx2; dx++) Sqrs[dz][dy][dx]=0.0;
	    }
	}

	StoreRange(Nx1,Nx2,dx1,dx2,&x11,&x12,&X21,&X22); nx=x12-x11+1;
	StoreRange(Ny1,Ny2,dy1,dy2,&y11,&y12,&Y21,&Y22); ny=y12-y11+1;
	StoreRange(Nz1,Nz2,dz1,dz2,&z11,&z12,&Z21,&Z22);

	if ((Value1=ALLOC(Value *,ny))==NULL) Error(vmae);

	for (y1=y11; y1<=y12; y1++) {
	    if ((Value1[y1]=ALLOC(Value,nx))==NULL) Error(vmae);

	    Value1[y1]-=x11;
	}

	x21=X21[x11]; x22=X22[x12]; nx=x22-x21+1;
	y21=Y21[y11]; y22=Y22[y12]; ny=y22-y21+1;

	if ((Value2=ALLOC(Value **,Dz))==NULL) Error(vmae);

	for (dz=0; dz<Dz; dz++) {
	    if ((Value2[dz]=ALLOC(Value *,ny))==NULL) Error(vmae);

	    Value2[dz]-=y21;

	    for (y2=y21; y2<=y22; y2++) {
		if ((Value2[dz][y2]=ALLOC(Value,nx))==NULL) Error(vmae);

		Value2[dz][y2]-=x21;
	    }
	}

	z21=Z21[z11]; z22=z21-1;

	for (z1=z11, Sz=Sqrs-z1; z1<=z12; z1++, Sz--) {
	    ScanValue(pbs1+z1,x11,y11,x12,y12,Value1);
	    for (z2=Z21[z1], sz=Sz+z2, Z2=Z22[z1]; z2<=Z2; z2++, sz++) {
		value2=Value2[(z2-z21)%Dz];
		if (z2>z22) {
		    z22=z2;
		    ScanValue(pbs2+z2,x21,y21,x22,y22,value2);
		}
		for (y1=y11, v1y=Value1+y1, Sy=(*sz)-y1;
		     y1<=y12; y1++, v1y++, Sy--) {
		    V1x=(*v1y);
		    for (y2=Y21[y1], v2y=value2+y2, sy=Sy+y2,
			 Y2=Y22[y1]; y2<=Y2; y2++, v2y++, sy++) {
			V2x=(*v2y);
			for (x1=x11, v1x=V1x+x1, Sx=(*sy)-x1;
			     x1<=x12; x1++, v1x++, Sx--) {
			    value1=(*v1x);
			    for (x2=X21[x1], v2x=V2x+x2, sx=Sx+x2,
				 X2=X22[x1]; x2<=X2; x2++, v2x++, sx++) {
				diff=value1-(*v2x); *sx+=(diff*diff);
			    }
			}
		    }
		}
	    }
	}

	for (dz=dz1; dz<=dz2; dz++) {
	    nz=min(z12+dz,z22)-max(z11+dz,z21)+1;
	    for (dy=dy1; dy<=dy2; dy++) {
		ny=min(y12+dy,y22)-max(y11+dy,y21)+1;
		for (dx=dx1; dx<=dx2; dx++) {
		    nx=min(x12+dx,x22)-max(x11+dx,x21)+1;

		    (void)printf("%d\t%d\t%d\t%lf\n",
				 dx,dy,dz,sqrt(Sqrs[dz][dy][dx]/((double)nx*
								 (double)ny*
								 (double)nz)));
		}
	    }
	}

	return 0;
}
