
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "cell.h"
#include "rif.h"

static void	Error(msg)
char		*msg;
{
	(void)fputs(msg,stderr); (void)fputc('\n',stderr); exit(1);
}

#define ALLOC(type,size)	(type *)malloc(sizeof(type)*(size))

static int	max(a,b)
int		a,b;
{
	return (a>b)?a:b;
}

#define PATH_LEN	256

static char	**CheckPathNxyz(Nx,Ny,Nz)
int		*Nx,*Ny,*Nz;
{
	char	**Path,path[PATH_LEN],*pmae="path memory allocation error.";
	int	z,nx,ny;

	if (scanf("%d",Nz)!=1 || *Nz<=0) Error("bad slice number.");

	if ((Path=ALLOC(char *,*Nz))==NULL) Error(pmae);

	*Nx=(*Ny)=0;
	for (z=0; z<*Nz; z++) {
	    if (scanf("%s",path)!=1) Error("path read error.");

	    ReadImageFile(path,&nx,&ny,NULL,NULL,NULL);
	    *Nx=max(*Nx,nx);
	    *Ny=max(*Ny,ny);

	    if ((Path[z]=ALLOC(char,strlen(path)+1))==NULL) Error(pmae);

	    (void)strcpy(Path[z],path);
	}
	return Path;
}

#define Value	Cell

static void	ScanValue(path,x1,y1,x2,y2,value)
char		*path;
int		x1,y1,x2,y2;
Value		**value;
{
	int	nx,ny,x,y;
	Cell	**cell;

	ReadImageFile(path,&nx,&ny,NULL,&cell,NULL);

	for (y=y1; y<=y2; y++)
	for (x=x1; x<=x2; x++) value[y][x]=(x<nx && y<ny)?(Value)cell[y][x]:0;

	while (--ny>=0) free(cell[ny]); free(cell);
}

static int	min(a,b)
int		a,b;
{
	return (a<b)?a:b;
}

static void	StoreRange(Nz1,Nz2,dz1,dz2,z11,z12,Z21,Z22)
int		Nz1,Nz2,dz1,dz2,*z11,*z12,**Z21,**Z22;
{
	int	nz,z1;

	*z11=max(0,-dz2); *z12=min(Nz1,Nz2-dz1)-1; nz=(*z12)-(*z11)+1;

	if ((*Z21=ALLOC(int,nz))==NULL ||
	    (*Z22=ALLOC(int,nz))==NULL) Error("range memory allocation error.");

	*Z21-=(*z11); *Z22-=(*z11);

	for (z1=(*z11); z1<=(*z12); z1++) {
	    (*Z21)[z1]=max(0,z1+dz1); (*Z22)[z1]=min(Nz2-1,z1+dz2);
	}
}

typedef struct {
		int	pels;
		double	sqrs;
	} Count;

int	main(argc,argv)
int	argc;
char	**argv;
{
	char	**path1,**path2,
		*cmae="count memory allocation error.",
		*vmae="value memory allocation error.";
	int	Nx1,Ny1,Nz1,Nx2,Ny2,Nz2,dx1,dy1,dz1,dx2,dy2,dz2,
		nx,ny,Dz,dz,dy,dx,
		x11,x12,*X21,*X22,y11,y12,*Y21,*Y22,z11,z12,*Z21,*Z22,
		y1,x21,x22,y21,y22,y2,z21,z22,z1,z2,Z2,Y2,x1,x2,X2;
	Count	***count,***Cz,***cz,**Cy,**cy,*Cx,*cx;
	Value	**Value1,***Value2,**value2,**v1y,*V1x,**v2y,*V2x,*v1x,*v2x;
	double	diff;

	if (argc!=7) Error("usage : dti3 dx1 dy1 dz1 dx2 dy2 dz2");

	path1=CheckPathNxyz(&Nx1,&Ny1,&Nz1);
	path2=CheckPathNxyz(&Nx2,&Ny2,&Nz2);

	if ((dx1=atoi(argv[1]))<=(-Nx1) ||
	    (dx2=atoi(argv[4]))>=  Nx2  || dx1>dx2 ||
	    (dy1=atoi(argv[2]))<=(-Ny1) ||
	    (dy2=atoi(argv[5]))>=  Ny2  || dy1>dy2 ||
	    (dz1=atoi(argv[3]))<=(-Nz1) ||
	    (dz2=atoi(argv[6]))>=  Nz2  || dz1>dz2)
	    Error("bad range of distance.");

	nx=dx2-dx1+1; ny=dy2-dy1+1; Dz=dz2-dz1+1;

	if ((count=ALLOC(Count **,Dz))==NULL) Error(cmae);

	count-=dz1;

	for (dz=dz1; dz<=dz2; dz++) {
	    if ((count[dz]=ALLOC(Count *,ny))==NULL) Error(cmae);

	    count[dz]-=dy1;

	    for (dy=dy1; dy<=dy2; dy++) {
		if ((count[dz][dy]=ALLOC(Count,nx))==NULL) Error(cmae);

		count[dz][dy]-=dx1;

		for (dx=dx1; dx<=dx2; dx++) {
		    count[dz][dy][dx].pels=0; count[dz][dy][dx].sqrs=0.0;
		}
	    }
	}

	StoreRange(Nx1,Nx2,dx1,dx2,&x11,&x12,&X21,&X22); nx=x12-x11+1;
	StoreRange(Ny1,Ny2,dy1,dy2,&y11,&y12,&Y21,&Y22); ny=y12-y11+1;
	StoreRange(Nz1,Nz2,dz1,dz2,&z11,&z12,&Z21,&Z22);

	if ((Value1=ALLOC(Value *,ny))==NULL) Error(vmae);

	for (y1=y11; y1<=y12; y1++) {
	    if ((Value1[y1]=ALLOC(Value,nx))==NULL) Error(vmae);

	    Value1[y1]-=x11;
	}

	x21=X21[x11]; x22=X22[x12]; nx=x22-x21+1;
	y21=Y21[y11]; y22=Y22[y12]; ny=y22-y21+1;

	if ((Value2=ALLOC(Value **,Dz))==NULL) Error(vmae);

	for (dz=0; dz<Dz; dz++) {
	    if ((Value2[dz]=ALLOC(Value *,ny))==NULL) Error(vmae);

	    Value2[dz]-=y21;

	    for (y2=y21; y2<=y22; y2++) {
		if ((Value2[dz][y2]=ALLOC(Value,nx))==NULL) Error(vmae);

		Value2[dz][y2]-=x21;
	    }
	}

	z21=Z21[z11]; z22=z21-1;

	for (z1=z11, Cz=count-z1; z1<=z12; z1++, Cz--) {
	    ScanValue(path1[z1],x11,y11,x12,y12,Value1);
	    for (z2=Z21[z1], cz=Cz+z2, Z2=Z22[z1]; z2<=Z2; z2++, cz++) {
		value2=Value2[(z2-z21)%Dz];
		if (z2>z22) {
		    z22=z2;
		    ScanValue(path2[z2],x21,y21,x22,y22,value2);
		}
		for (y1=y11, v1y=Value1+y1, Cy=(*cz)-y1;
		     y1<=y12; y1++, v1y++, Cy--) {
		    V1x=(*v1y);
		    for (y2=Y21[y1], v2y=value2+y2, cy=Cy+y2,
			 Y2=Y22[y1]; y2<=Y2; y2++, v2y++, cy++) {
			V2x=(*v2y);
			for (x1=x11, v1x=V1x+x1, Cx=(*cy)-x1;
			     x1<=x12; x1++, v1x++, Cx--)
			    if (*v1x)
				for (x2=X21[x1], v2x=V2x+x2, cx=Cx+x2,
				     X2=X22[x1]; x2<=X2; x2++, v2x++, cx++)
				    if (*v2x) {
					++cx->pels;
					diff=(double)*v1x-(double)*v2x;
					cx->sqrs+=(diff*diff);
				    }
		    }
		}
	    }
	}

	for (dz=dz1; dz<=dz2; dz++)
	for (dy=dy1; dy<=dy2; dy++)
	for (dx=dx1; dx<=dx2; dx++)
	    if (count[dz][dy][dx].pels)
		(void)printf("%d\t%d\t%d\t%lf\n",dx,dy,dz,
		sqrt(count[dz][dy][dx].sqrs/(double)count[dz][dy][dx].pels));

	return 0;
}
