
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

#define MA(noe,ptr)	malloc((noe)*sizeof(*(ptr)))

typedef struct Edge {
		int		nop;
		double		*x,*y;
		struct Edge	*next;
	} Edge;

static void	ReadEdge(xC,yC,dx,dy,edge,total,radius)
double		xC,yC,dx,dy,*radius;
Edge		**edge;
int		*total;
{
	int	nop;
	double	r,s,*x,*y,rs1,rs2=0.0;
	char	*bed="bad edge data.";
	Edge	*e;

	*total=0;
	while (scanf("%d",&nop)==1 && nop!=0) {
	    if (nop<0) nop=(-nop);

	    if (nop==1) {
		if (scanf("%lf %lf",&r,&s)!=2) Error(bed);

		continue;
	    }

	    if ((e=(Edge   *)MA(1,  e))==NULL ||
		(x=(double *)MA(nop,x))==NULL ||
		(y=(double *)MA(nop,y))==NULL)
		Error("edge data memory allocation error.");

	    *total+=nop;
	    *edge=e; edge=(&(e->next));
	    e->nop=nop; e->x=x; e->y=y; e->next=NULL;

	    do {
		if (scanf("%lf %lf",x,y)!=2) Error(bed);

		*x=r=dx*(*x-xC); ++x;
		*y=s=dy*(*y-yC); ++y; if ((rs1=r*r+s*s)>rs2) rs2=rs1;
	    } while (--nop>0);
	}
	if (*total==0) Error("no edge data.");

	*radius=sqrt(rs2);
}

typedef struct {
		union {
			long	l;
			double	d;
		} r;
		int	nL,nR;
		double	*sL,*sR,tL,tR;
	} Cut;

#define RL	r.l
#define RD	r.d

#define LONG_MIN	(-2147483647L)
#define LONG_MAX	2147483647L
#define LONG_SPAN	((double)LONG_MAX-(double)LONG_MIN)

static void	PrepareCut(total,radius,N,dr,r0,cut,base,unit)
int		total,N;
double		radius,dr,r0,*base,*unit;
Cut		**cut;
{
	double	r1,rN1,r2;

	if ((*cut=(Cut *)MA(total+2,*cut))==NULL)
	    Error("cut data memory allocation error.");

	r1=(r0<-radius)?r0:-radius;
	rN1=r0+dr*(double)(N-1); r2=(rN1>radius)?rN1:radius;
	*base=(r1*(double)LONG_MAX-r2*((double)LONG_MIN+1.0))/(LONG_SPAN-1.0);
	*unit=(LONG_SPAN-3.0)/(r2-r1+dr*2.0);
}

#define RD2RL(rd)	(long)floor(((rd)-base)*unit)
#define RL2RD(rl)	(base+((double)(rl)+0.5)/unit)

static void	RotateEdge(edge,angle)
Edge		*edge;
double		angle;
{
	double	*x,*y,r,s,Cos=cos(angle),Sin=sin(angle);
	int	nop;

	do {
	    x=edge->x; y=edge->y;
	    for (nop=edge->nop; nop>0; nop--) {
		r=(*x); s=(*y); *(x++)=r*Cos+s*Sin; *(y++)=s*Cos-r*Sin;
	    }
	} while ((edge=edge->next)!=NULL);
}

#define NL	(c->nL)
#define NR	(c->nR)
#define SL	(c->sL)
#define SR	(c->sR)

static int	CompareCutR(i,j)
Cut		*i,*j;
{
	return (i->RL<j->RL)?-1:(i->RL>j->RL)?1:0;
}

static int	CompareCutS(i,j)
double		*i,*j;
{
	return (*i<*j)?-1:(*i>*j)?1:0;
}

static void	StoreCut(edge,cut,total,base,unit)
Edge		*edge;
Cut		*cut;
int		total;
double		base,unit;
{
	Cut	*c;
	Edge	*e;
	double	*x,*y,s1,s2,dsdr;
	int	nop,count,index,d1,d2,n;
	long	r1,r2;

	for (c=cut+1, e=edge; e!=NULL; e=e->next)
	    for (x=e->x, nop=e->nop; nop>0; nop--)
		(c++)->RL=RD2RL(*(x++));

	qsort(cut+1,total,sizeof(*cut),CompareCutR);

	cut->RL=LONG_MIN; count=1;

	for (c=cut+2, index=1; index<total; index++, c++)
	    if (c->RL>cut[count].RL) cut[++count].RL=c->RL;

	cut[++count].RL=LONG_MAX; ++count;

	for (c=cut, index=0; index<count; index++, c++) NL=NR=0;

	for (e=edge; e!=NULL; e=e->next) {
	    nop=e->nop; x=e->x;

	    r1=RD2RL(x[nop-2]);
	    r2=RD2RL(x[nop-1]); d2=(r1>r2)?-1:(r1==r2)?0:1;

	    for (c=cut+1; c->RL<r2; c++) ;

	    do {
		r1=r2; r2=RD2RL(*(x++));
		d1=d2; d2=(r1>r2)?-1:(r1==r2)?0:1;

		switch(d1*3+d2+4) {
/* -1 & -1 */	    case 0 : ++NL; ++NR; break;
/* -1 &  0 */	    case 1 : ++NR; break;
/* -1 & +1 */	    case 2 : NR+=2; break;
/*  0 & -1 */	    case 3 : ++NL; break;
/*  0 &  0 */	    case 4 : break;
/*  0 & +1 */	    case 5 : ++NR; break;
/* +1 & -1 */	    case 6 : NL+=2; break;
/* +1 &  0 */	    case 7 : ++NL; break;
/* +1 & +1 */	    case 8 : ++NL; ++NR; break;
		}
		if (d2!=0)
		    while ((c+=d2)->RL!=r2) {
			++NL; ++NR;
		    }
	    } while (--nop>0);
	}

	for (c=cut, index=0; index<count; index++, c++) {
	    if ((NL%2)!=0 || (NR%2)!=0) Error("bad intersection number.");

	    if ((NL!=0 && (SL=(double *)MA(NL,SL))==NULL) ||
		(NR!=0 && (SR=(double *)MA(NR,SR))==NULL))
		Error("intersection data memory allocation error.");

	    NL=NR=0;
	}

	for (e=edge; e!=NULL; e=e->next) {
	    nop=e->nop; x=e->x; y=e->y;

	    r1=RD2RL(x[nop-2]);
	    r2=RD2RL(x[nop-1]); d2=(r1>r2)?-1:(r1==r2)?0:1;
	    s2=y[nop-1];

	    for (c=cut+1; c->RL<r2; c++) ;

	    do {
		r1=r2; r2=RD2RL(*(x++));
		d1=d2; d2=(r1>r2)?-1:(r1==r2)?0:1;
		s1=s2; s2=(*(y++));

		switch(d1*3+d2+4) {
/* -1 & -1 */	    case 0 : SL[NL++]=SR[NR++]=s1; break;
/* -1 &  0 */	    case 1 : SR[NR++]=s1; break;
/* -1 & +1 */	    case 2 : SR[NR++]=SR[NR++]=s1; break;
/*  0 & -1 */	    case 3 : SL[NL++]=s1; break;
/*  0 &  0 */	    case 4 : break;
/*  0 & +1 */	    case 5 : SR[NR++]=s1; break;
/* +1 & -1 */	    case 6 : SL[NL++]=SL[NL++]=s1; break;
/* +1 &  0 */	    case 7 : SL[NL++]=s1; break;
/* +1 & +1 */	    case 8 : SL[NL++]=SR[NR++]=s1; break;
		}
		if (d2!=0) {
		    dsdr=(s2-s1)/((double)r2-(double)r1);
		    while ((c+=d2)->RL!=r2)
			SL[NL++]=SR[NR++]=s1+dsdr*((double)c->RL-(double)r1);
		}
	    } while (--nop>0);
	}

	for (c=cut, index=0; index<count; index++, c++) {
	    c->tL=0.0;
	    if (NL!=0) {
		qsort(SL,NL,sizeof(*SL),CompareCutS);

		for (n=0; n<NL; n+=2) c->tL+=(SL[n+1]-SL[n]);

		free(SL);
	    }
	    c->tR=0.0;
	    if (NR!=0) {
		qsort(SR,NR,sizeof(*SR),CompareCutS);

		for (n=0; n<NR; n+=2) c->tR+=(SR[n+1]-SR[n]);

		free(SR);
	    }
	    c->RD=RL2RD(c->RL);
	}
}

#define Thickness(r12)	c->tR+(cut->tL-c->tR)*(r12-c->RD)/(cut->RD-c->RD)

static void	PrintThickness(cut,N,dr,r0)
Cut		*cut;
int		N;
double		dr,r0;
{
	int	n;
	Cut	*c=cut-1;

#ifdef	POINT_DETECTOR
	double	r;

	for (n=0; n<N; n++) {
	    for (r=r0+dr*(double)n; cut->RD<r; c=cut++) ;

	    printf("%lf\n",Thickness(r));
	}
#else
	double	r1,r2,sum;

	r0-=(dr*0.5);

	for (r1=r0; cut->RD<r1; c=cut++) ;

	for (n=0; n<N; n++, r1=r2) {
	    sum=0.0;
	    for (r2=r0+dr*(double)n; cut->RD<r2; r1=cut->RD, c=cut++)
		sum+=((Thickness(r1)+cut->tL)*(cut->RD-r1));

	    printf("%lf\n",(sum+(Thickness(r1)+Thickness(r2))*(r2-r1))/dr);
	}
#endif
}

#ifndef M_PI
#define M_PI	3.14159265358979323846
#endif

int	main(argc,argv)
int	argc;
char	**argv;
{
	Edge	*edge;
	int	total,N,M,m;
	double	radius,dr,r0,base,unit,theta0,dtheta;
	Cut	*cut;

	if (argc!=10) Error("usage : e2t xC yC dx dy N dr r0/dr M theta0");

	ReadEdge(atof(argv[1]),atof(argv[2]),
		 atof(argv[3]),atof(argv[4]),&edge,&total,&radius);

	if ((N=atoi(argv[5]))<=0) Error("bad detector number.");

	if ((dr=atof(argv[6]))<=0.0) Error("bad detector interval.");

	r0=atof(argv[7])*dr;

	PrepareCut(total,radius,N,dr,r0,&cut,&base,&unit);

	if ((M=atoi(argv[8]))<=0) Error("bad rotation step number.");

	theta0=atof(argv[9])*M_PI/180.0;

	dtheta=M_PI/(double)M;

	printf("%s\t%s\t%lf\n%s\t%s\n",argv[5],argv[6],r0,argv[8],argv[9]);

	for (m=0; m<M; m++) {
	    RotateEdge(edge,(m==0)?theta0:dtheta);

	    StoreCut(edge,cut,total,base,unit);

	    PrintThickness(cut,N,dr,r0);
	}
	return 0;
}
