
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include "FBP.h"

/*----------------------------------------------------------------------*/

#define R_BASE	28.0
#define R_UNIT	0.005

#define D_BASE	128.0
#define D_UNIT	0.01

/*----------------------------------------------------------------------*/

static int	CheckPixel(data)
double		data;
{
	return (data<0.0)?0:(data>255.0)?255:(int)(data+0.5);
}

static FILE	*imageFile;

#define ImageNotStored(name,RDImage)	\
{\
	fprintf(stderr,"%s : %s image not stored.\n",name,RDImage); return 0;\
}

/*----------------------------------------------------------------------*/

#ifdef	BMP

#define REAL_SIZE	0.2	/* 0.2 meter */

static void	PutLong(l)
long		l;
{
	fputc((int)((l    )&0xffL),imageFile);
	fputc((int)((l>> 8)&0xffL),imageFile);
	fputc((int)((l>>16)&0xffL),imageFile);
	fputc((int)((l>>24)&0xffL),imageFile);
}

static void	PutShort(s)
short		s;
{
	fputc((s   )&0xff,imageFile);
	fputc((s>>8)&0xff,imageFile);
}

static int	linePad;

static int	OpenImageFile(name,RDImage,N)
char		*name,*RDImage;
int		N;
{
	long	IDS,DPM;
	int	gray;

	if ((imageFile=fopen(name,"wb"))==NULL) ImageNotStored(name,RDImage);

	linePad=((N+3)/4)*4-N; IDS=(long)(((N+3)/4)*4)*(long)N;
			       DPM=(long)ceil((double)N/REAL_SIZE);

	fputc('B',imageFile);
	fputc('M',imageFile);
	PutLong(14L+40L+256L*4L+IDS);	/* file size */
	PutShort(0);			/* reserved */
	PutShort(0);			/* reserved */
	PutLong(14L+40L+256L*4L);	/* header size */
	PutLong(40L);			/* information size */
	PutLong((long)N);		/* image width */
	PutLong((long)N);		/* image height */
	PutShort(1);			/* image planes */
	PutShort(8);			/* image Bits Per Sample */
	PutLong(0L);			/* compression mode */
	PutLong(IDS);			/* image data size */
	PutLong(DPM);			/* Dots Per Meter along x */
	PutLong(DPM);			/* Dots Per Meter along y */
	PutLong(256L);			/* number of colors */
	PutLong(256L);			/* important colors */

	for (gray=0; gray<256; gray++) {
	    fputc(gray,imageFile);
	    fputc(gray,imageFile);
	    fputc(gray,imageFile); fputc(0,imageFile);
	}
	return 1;
}

#define ForV(N)	for (v=N-1; v>=0; v--)
#define ForH(N)	for (h=0; h<N; h++)

#define StorePixel(data)	fputc(CheckPixel(data),imageFile)

static void	EndOfScanLine()
{
	int	count=linePad;

	while (--count>=0) fputc(0,imageFile);
}

#define CloseImageFile()	fclose(imageFile)

/*----------------------------------------------------------------------*/

#else	/* ifndef BMP */

/*----------------------------------------------------------------------*/

#ifdef	PGMRAW

#define MAGIC			"P5"

#define StorePixel(data)	fputc(CheckPixel(data),imageFile)

#else	/* ifndef PGMRAW */

#define MAGIC			"P2"

#define StorePixel(data)	fprintf(imageFile,"%d\n",CheckPixel(data))

#endif	/* PGMRAW */

static int	OpenImageFile(name,RDImage,N)
char		*name,*RDImage;
int		N;
{
	if (*name=='-')
	    imageFile=stdout;
	else
	    if ((imageFile=fopen(name,"w"))==NULL) ImageNotStored(name,RDImage);

	fprintf(imageFile,"%s\n%d\t%d\t255\n",MAGIC,N,N);
	return 1;
}

#define ForV(N)			for (v=0; v<N; v++)
#define ForH(N)			for (h=0; h<N; h++)

#define EndOfScanLine()		/* nothing */

#define CloseImageFile()	if (imageFile!=stdout) fclose(imageFile)

#endif	/* BMP */

/*----------------------------------------------------------------------*/

static void	StoreReconstructedImage(N,f,name)
int		N;
Float		**f;
char		*name;
{
	int	v,h;

	if (OpenImageFile(name,"reconstructed",N)) {
	    ForV(N) {
		ForH(N) { StorePixel(R_BASE+f[v][h]/R_UNIT); }
		EndOfScanLine();
	    }
	    CloseImageFile();
	}
}

static void	StoreDifferentialImage(N,f,delta,x0,y0,xC,yC,radius,name)
int		N;
Float		**f;
double		delta,x0,y0,xC,yC,radius;
char		*name;
{
	int	v,h;
	double	dx2,dy2,df,mean,min=0.0,sum1=0.0,
				max=0.0,sum2=0.0,sum0=(double)N*(double)N,
						 radius2=radius*radius;

	if (OpenImageFile(name,"differential",N)) {
	    ForV(N) {
		dy2=y0-delta*(double)v-yC; dy2*=dy2;
		ForH(N) {
		    dx2=x0+delta*(double)h-xC; dx2*=dx2;
		    df=(dx2+dy2<=radius2)?f[v][h]-1.0:f[v][h];
		    StorePixel(D_BASE+df/D_UNIT);

		    if (df<min) min=df;
		    if (df>max) max=df; sum1+=df; sum2+=(df*df);
		}
		EndOfScanLine();
	    }
	    CloseImageFile();

	    fprintf(stderr,"differences min.\t%lf\n",min);
	    fprintf(stderr,"differences max.\t%lf\n",max);

	    mean=sum1/sum0;
	    fprintf(stderr,"differences mean\t%lf\n",mean);
	    fprintf(stderr,"differences S.D.\t%lf\n",sqrt(sum2/sum0-mean*mean));
	}
}

/*----------------------------------------------------------------------*/

static void	StoreProjection(fbp,r0,xC,yC,radius)
FBP		*fbp;
double		r0,xC,yC,radius;
{
	int	m,n,M=fbp->M,N=fbp->N;
	double	rC,r,s2,*Cos=fbp->Cos,
			*Sin=fbp->Sin,delta=fbp->delta,radius2=radius*radius;
	Float	**p=fbp->pq;

	for (m=0; m<M; m++) {
	    rC=xC*Cos[m]+yC*Sin[m];
	    for (n=0; n<N; n++) {
		r=r0+delta*(double)n;
		s2=radius2-(r-rC)*(r-rC);
		p[m][n]=(s2>0.0)?2.0*sqrt(s2):0.0;
	    }
	}
}

/*----------------------------------------------------------------------*/

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

/*----------------------------------------------------------------------*/

#ifndef	CLOCKS_PER_SEC
#define CLOCKS_PER_SEC	1000000

extern long	clock();
#endif

#define CLOCK()	((double)clock()/(double)CLOCKS_PER_SEC)

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	N,M,regriding=0;
	double	N12,xC,yC,radius,Clock;
	FBP	*fbp;
	Float	**f;

	if (argc!=7 && argc!=8)
	    Error("usage : fbptst N {M} xCenter yCenter radius RImage DImage");

	if ((N=atoi(argv[1]))<=0) Error("bad N.");

	N12=(double)(N-1)/2.0;

	if (argc==7) {
	    M=(int)ceil(M_PI*(double)N/2.0);

	    fprintf(stderr,"M\t%d\n",M);
	}
	else {
	    if ((M=atoi(argv[2]))==0) Error("bad M.");

	    if (M<0) {
		M=(-M); regriding=1;
	    }
	}

	Clock=CLOCK();

	if ((fbp=SetUpFBP(N,M,1.0,-N12,0.0))==NULL)
	    Error("memory allocation error for FBP.");

	fprintf(stderr,"SetUpFBP\t%lf\n",CLOCK()-Clock);

	xC=atof(argv[argc-5]);
	yC=atof(argv[argc-4]);
	radius=atof(argv[argc-3]);

	StoreProjection(fbp,-N12,xC,yC,radius);

	Clock=CLOCK();

	Filtering(fbp);

	fprintf(stderr,"Filtering\t%lf\n",CLOCK()-Clock);

	if (regriding)
{
	Clock=CLOCK();

	if (!ReGridingFP(fbp)) Error("memory allocation error for FP.");

	fprintf(stderr,"ReGridingFP\t%lf\n",CLOCK()-Clock);

	fprintf(stderr,"M\t%d\t%d\n",M,fbp->M);
}

#ifdef	BPIMAGE

	Clock=CLOCK();

	if ((f=BPImage(fbp))==NULL)
	    Error("memory allocation error for BPImage.");

	fprintf(stderr,"BPImage\t%lf\n",CLOCK()-Clock);

#else	/* ifndef BPIMAGE */
{
	int	v,h;

	if ((f=(Float **)malloc(N*sizeof(*f)))==NULL)
	    Error("memory allocation error for BackProjection (1).");

	for (v=0; v<N; v++)
	    if ((f[v]=(Float *)malloc(N*sizeof(**f)))==NULL)
		Error("memory allocation error for BackProjection (2).");

	Clock=CLOCK();

	for (v=0; v<N; v++) for (h=0; h<N; h++)
	    f[v][h]=BackProjection(fbp,(double)h-N12,N12-(double)v);

	fprintf(stderr,"BackProjection\t%lf\n",CLOCK()-Clock);
}
#endif	/* BPIMAGE */

	StoreReconstructedImage(N,f,argv[argc-2]);
	StoreDifferentialImage(N,f,1.0,-N12,N12,xC,yC,radius,argv[argc-1]);
	return 0;
}
