
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>

#define time	tick	/* to prevent the error of lint */

#define Error(msg)	{ fprintf(stderr,"%s\n",msg); exit(1); }
#define Abort(msg)	{ fprintf(stderr,"%s : %s\n",path,msg); exit(1); }
#define MA(cnt,ptr)	malloc((cnt)*sizeof(*(ptr)))

#define GetByte(file)	fgetc(file)

static int	GetWord(file)
FILE		*file;
{
	int	lo,hi;

	return ((lo=GetByte(file))==EOF ||
		(hi=GetByte(file))==EOF)?-1:(hi<<8)|lo;
}

#define DATE_LEN	256

static double	GetTime(path,len,file)
char		*path;
int		len;
FILE		*file;
{
	int		code,comma=0;
	char		date[DATE_LEN],*ptr=date;
	struct tm	tm;

	while (--len>=0)
	    if (GetByte(file)==',' && ++comma==11) break;

	if (comma!=11) Abort("top of imaging date not found.");

	while (--len>=0) {
	    if ((code=GetByte(file))==',')
		if (++comma==13)
		    break;
		else
		    code=' ';
	    else
		if (code=='-' || code==':') code=' ';

	    *(ptr++)=code;
	}

	if (comma!=13) Abort("end of imaging date not found.");

	*ptr='\0';

	if (sscanf(date,"%d %d %d %d %d %d",&(tm.tm_year),
					    &(tm.tm_mon ),
					    &(tm.tm_mday),
					    &(tm.tm_hour),
					    &(tm.tm_min ),
					    &(tm.tm_sec ))!=6)
	    Abort("bad format of imaging date.");

	while (--len>=0) (void)GetByte(file);

	tm.tm_year-=1900; tm.tm_mon-=1; return (double)mktime(&tm);
}

static int	Nx,Ny,Nxy,firstImage=1;

static void	ReadImage(path,time,line)
char		*path;
double		*time,**line;
{
	FILE	*file=fopen(path,"rb");
	int	len,n,z;

	if (file==NULL) Abort("file not found.");

	if (GetByte(file)!='I' ||
	    GetByte(file)!='M') Abort("bad magic number.");

	if ((len=GetWord(file))<0) Abort("bad comment length.");

	if (firstImage) {
	    firstImage=0;

	    if ((Nx=GetWord(file))<=0 ||
		(Ny=GetWord(file))<=0) Abort("bad image size.");

	    Nxy=Nx*Ny;
	}
	else
	    if (GetWord(file)!=Nx ||
		GetWord(file)!=Ny) Abort("image size not match.");

	if ((*line=(double *)MA(Nxy,*line))==NULL)
	    Abort("scanline memory allocation error.");

	fseek(file,4L,1);	/* image offset */

	if (GetWord(file)!=2) Abort("bad image type.");

	fseek(file,50L,1);	/* reserved data */

	if (time==NULL)
	    fseek(file,(long)len,1);
	else
	    *time=GetTime(path,len,file);

	for (n=0; n<Nxy; n++) {
	    if ((z=GetWord(file))<0) Abort("unexpected end of file.");

	    (*line)[n]=(double)z;
	}
	fclose(file);
}

#define TEXT_LEN	512
#define PATH_LEN	256
#define COMMENT_CODE	'#'

static int	GetItem(path,name,time,file)
char		*path,*name;
double		*time;
FILE		*file;
{
	char	text[TEXT_LEN],*ptr;
	int	cnt=0;

	while (fgets(text,TEXT_LEN,file)!=NULL) {
	    if (text[strlen(text)-1]!='\n') Abort("too long line.");

	    if ((ptr=strchr(text,COMMENT_CODE))!=NULL) *ptr='\0';

	    if ((cnt=sscanf(text,"%s %lf",name,time))!=0) break;
	}
	return cnt;
}

static void	SortData(time,line,bgn,end)
double		*time,**line;
int		bgn,end;
{
	int	b=bgn,e=end;
	double	t,*l,key=time[(b+e)/2];

	do {
	    while (time[b]<key) ++b;
	    while (time[e]>key) --e;

	    if (b<=e) {
		if (b!=e) {
		    t=time[b]; time[b]=time[e]; time[e]=t;
		    l=line[b]; line[b]=line[e]; line[e]=l;
		}
		++b; --e;
	    }
	} while (b<=e);

	if (bgn<e) SortData(time,line,bgn,e);
	if (b<end) SortData(time,line,b,end);
}

static void	StoreMean(time,line,m0,m1,m2)
double		*time,**line;
int		m0,m1,m2;
{
	int	n,m,dm=m2-m1;

	if (m0!=m1) {
	    time[m0]=time[m1];
	    for (n=0; n<Nxy; n++) line[m0][n]=line[m1][n];
	}
	for (m=m1+1; m<m2; m++) {
	    time[m0]+=time[m];
	    for (n=0; n<Nxy; n++) line[m0][n]+=line[m][n];
	}
	if (dm!=1) {
	    time[m0]/=(double)dm;
	    for (n=0; n<Nxy; n++) line[m0][n]/=(double)dm;
	}
}

static int	firstTime=1;
static double	time1,time2;

static void	CompareTime(time)
double		time;
{
	if (firstTime) {
	    firstTime=0; time1=time2=time;
	}
	else {
	    if (time<time1) time1=time;
	    if (time>time2) time2=time;
	}
}

typedef struct {
		int	No;
		double	*time,**line;
	} Table;

#define TIME_UNIT	0.0001

static void	ReadTable(path,table)
char		*path;
Table		*table;
{
	FILE	*file;
	int	No,m,m1,m2,t1,t2;
	char	name[PATH_LEN];
	double	t,*time,**line;

	if (*path=='-')
	    No=1;
	else {
	    if ((file=fopen(path,"r"))==NULL) Abort("file not found.");

	    switch(GetItem(path,name,&t,file)) {
		case 0 :
		    Abort("unexpected end of file.");
		case 1 :
		    if (sscanf(name,"%d",&No)==1) {
			if (No<=0) Abort("bad number of lines.");

			break;
		    }
		case 2 :
		    for (No=1; GetItem(path,name,&t,file)!=0; No++) ;

		    fseek(file,0L,0); break;
	    }
	}

	m=(No==1)?2:No;
	if ((time=(double  *)MA(m,time))==NULL ||
	    (line=(double **)MA(m,line))==NULL)
	    Abort("table memory allocation error.");

	if (*path=='-') {
	    ReadImage(&(path[1]),time,line);
	    CompareTime(*time);
	}
	else {
	    for (m=0; m<No; m++) {
		ReadImage(name,(GetItem(path,name,&(time[m]),file)==2)?NULL:
			  &(time[m]),&(line[m]));
		CompareTime(time[m]);
	    }
	    fclose(file);
	}

	SortData(time,line,0,No-1);

	t=time[0]; m=m1=t1=0;
	for (m2=1; m2<No; m2++) {
	    t2=(int)floor((time[m2]-t)/TIME_UNIT);
	    if (t1!=t2) {
		StoreMean(time,line,m,m1,m2);
		++m; m1=m2; t1=t2;
	    }
	}
	StoreMean(time,line,m,m1,m2);
	table->No=m+1;

	table->time=time; table->line=line;
}

static void	WriteIntensity(path,d,t0,l0)
char		*path;
Table		*d;
double		t0,*l0;
{
	FILE	*file;
	int	k,n,No2=d->No-2;
	double	*t1,r1,r2,*l1,*l2,i;

	if (*path=='-')
	    file=stdout;
	else
	    if ((file=fopen(path,"wb"))==NULL) Abort("file not stored.");

	fprintf(file,"%d\t%d\n",Nx,Ny);

	for (t1=(&(d->time[k=No2])); k>0 && t0<*t1; k--, t1--) ;
	r1=(t0-*t1)/(t1[1]-*t1); r2=1.0-r1; l1=d->line[k]; l2=d->line[k+1];
	for (n=0; n<Nxy; n++) {
	    i=(*(l0++))-(r2*(*(l1++))+r1*(*(l2++)));
	    fwrite(&i,sizeof(i),1,file);
	}
	if (file!=stdout) fclose(file);
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	Table	d;
	double	time,*line;

	if (argc!=4 && argc!=5)
	    Error("usage : hp2iexy dark.tbl ie.img {ie_time} IE");

	ReadTable(argv[1],&d);

	if (argc==4)
	    ReadImage(argv[2],&time,&line);
	else {
	    time=atof(argv[3]);

	    ReadImage(argv[2],NULL,&line);
	}
	CompareTime(time);

	if (d.No==1) {
	    d.No=2;
	    d.time[0]=time1*2.0-time2;
	    d.time[1]=time2*2.0-time1;
	    d.line[1]=d.line[0];
	}
	WriteIntensity(argv[argc-1],&d,time,line);

	return 0;
}
