
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

static FILE	*hp;

#define GetByte()	fgetc(hp)

static int	GetWord()
{
	int	lo,hi;

	return ((lo=GetByte())==EOF || (hi=GetByte())==EOF)?-1:(hi<<8)|lo;
}

#define DATE_LEN	256

static time_t	GetTime(len)
int		len;
{
	int		code,comma=0;
	char		date[DATE_LEN],*ptr=date;
	struct tm	tm;
	time_t		t=(time_t)-1;

	while (--len>=0) if (GetByte()==',' && ++comma==11) break;

	if (comma==11) {
	    while (--len>=0) {
		if ((code=GetByte())==',')
		    if (++comma==13)
			break;
		    else
			code=' ';
		else
		    if (code=='-' || code==':') code=' ';

		*(ptr++)=code;
	    }
	    if (comma==13) {
		*ptr='\0';

		if (sscanf(date,"%d %d %d %d %d %d",&(tm.tm_year),
						    &(tm.tm_mon ),
						    &(tm.tm_mday),
						    &(tm.tm_hour),
						    &(tm.tm_min ),
						    &(tm.tm_sec ))==6) {
		    tm.tm_year-=1900; tm.tm_mon-=1; t=mktime(&tm);
		}
	    }
	}
        while (--len>=0) (void)GetByte();

	return t;
}

#define PutByte(data)	fputc(data,hp)

static void	PutWord(data)
int		data;
{
	PutByte(data&0xff); PutByte((data>>8)&0xff);
}

#define COMMENT_LEN	101

static void	PutTime(t)
time_t		t;
{
	int		comma,idx;
	char		date[DATE_LEN];
	struct tm	*tm=localtime(&t);

	for (comma=0; comma<11; comma++) PutByte(',');

	sprintf(date,"%04d-%02d-%02d,%02d:%02d:%02d,",tm->tm_year+1900,
						      tm->tm_mon+1,
						      tm->tm_mday,
						      tm->tm_hour,
						      tm->tm_min,
						      tm->tm_sec);
	for (idx=0; idx<20; idx++) PutByte(date[idx]);

	for (comma=13; comma<83; comma++) PutByte(',');
}

#define Error(msg)	{ fprintf(stderr,"%s\n",msg); exit(1); }
#define Abort(msg)	{ fprintf(stderr,"%s : %s\n",*argv,msg); exit(1); }

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	arg,len,Nx,Ny,x,y,z,idx,timeExists=1;
	double	*cell,*ptr,No,timer=0.0;
	time_t	t;

	if (argc<3) Error("usage : hp2mean HiPic {Hipic ...} meanHiPic");

	for (arg=1; arg<argc-1; arg++) {
	    if ((hp=fopen(*(++argv),"rb"))==NULL) Abort("file not found.");

	    if (GetByte()!='I' ||
		GetByte()!='M') Abort("bad magic number.");

	    if ((len=GetWord())<0) Abort("bad comment length.");

	    if (arg==1) {
		if ((Nx=GetWord())<=0 ||
		    (Ny=GetWord())<=0) Abort("bad image size.");

		if ((cell=(double *)malloc(Nx*Ny*sizeof(*cell)))==NULL)
		    Error("cell memory allocation error.");

		ptr=cell; for (y=0; y<Ny; y++)
			  for (x=0; x<Nx; x++) *(ptr++)=0.0;
	    }
	    else
		if (GetWord()!=Nx ||
		    GetWord()!=Ny) Abort("image size not match.");

	    fseek(hp,4L,1);	/* image offset */

	    if (GetWord()!=2) Abort("bad image type.");

	    fseek(hp,50L,1);	/* reserved data */

	    if ((t=GetTime(len))==(time_t)-1) {
		fprintf(stderr,"%s : time data not found (warning).\n",*argv);

		timeExists=0;
	    }
	    else
		timer+=(double)t;

	    ptr=cell;
	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++) {
		if ((z=GetWord())<0) Abort("unexpected end of file.");

		*(ptr++)+=(double)z;
	    }
	    fclose(hp);
	}
	No=(double)(argc-2);

	if (**(++argv)=='-')
	    hp=stdout;
	else
	    if ((hp=fopen(*argv,"wb"))==NULL) Abort("file not stored.");

	PutByte('I');	/* magic number */
	PutByte('M');

	PutWord((timeExists)?COMMENT_LEN:0);

	PutWord(Nx);	/* image size */
	PutWord(Ny);
	PutWord(0);	/* image offset */
	PutWord(0);
	PutWord(2);	/* image type */

	for (idx=0; idx<50; idx++) PutByte(0);	/* reserved data */

	if (timeExists) PutTime((time_t)(timer/No+0.5));

	ptr=cell; for (y=0; y<Ny; y++)
		  for (x=0; x<Nx; x++) PutWord((int)(*(ptr++)/No+0.5));

	if (hp!=stdout) fclose(hp);

	return 0;
}
