
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "sif.h"

static FILE	*hp;

#define Get1()	fgetc(hp)

static int	Get2()
{
	int	lo,hi;

	return ((lo=Get1())==EOF || (hi=Get1())==EOF)?-1:(hi<<8)|lo;
}

#define ROUND(d)	(Cell)(d+0.5)

#ifdef	FUNCTION_CELL
static double	**cell;

static Cell	GetCell(x,y)
int		x,y;
{
	return ROUND(cell[y][x]);
}
#endif

#define Error(msg)	{ fprintf(stderr,"%s\n",msg); exit(1); }
#define Abort(msg)	{ fprintf(stderr,"%s : %s\n",*argv,msg); exit(1); }
#define MA(cnt,ptr)	malloc((cnt)*sizeof(*(ptr)))

#define BPS	8*sizeof(Cell)

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	arg,len,Nx,Ny,x,y,z;
	double	**mean,**SD,c;
	char	*cmae="cell memory allocation error.";

	if (argc<4) Error("usage : hp2msd HiPic {Hipic ...} meanTIFF SDTIFF");

	for (arg=1; arg<argc-2; arg++) {
	    if ((hp=fopen(*(++argv),"rb"))==NULL) Abort("file not found.");

	    if (Get1()!='I' || Get1()!='M') Abort("bad magic number.");

	    if ((len=Get2())<0) Abort("bad comment length.");

	    if (arg==1) {
		if ((Nx=Get2())<=0 ||
		    (Ny=Get2())<=0) Abort("bad image size.");

		if ((mean=(double **)MA(Ny,mean))==NULL ||
		    (SD  =(double **)MA(Ny,SD  ))==NULL) Error(cmae);

		for (y=0; y<Ny; y++) {
		    if ((mean[y]=(double *)MA(Nx,*mean))==NULL ||
			(SD[y]  =(double *)MA(Nx,*SD  ))==NULL) Error(cmae);

		    for (x=0; x<Nx; x++) mean[y][x]=SD[y][x]=0.0;
		}
	    }
	    else
		if (Get2()!=Nx ||
		    Get2()!=Ny) Abort("image size not match.");

	    fseek(hp,(long)(12-8),1);		/* image offset */

	    if (Get2()!=2) Abort("bad image type.");

	    fseek(hp,(long)(64+len-14),1);	/* reserved data and comment */

	    for (y=0; y<Ny; y++) for (x=0; x<Nx; x++) {
		if ((z=Get2())<0) Abort("unexpected end of file.");

		mean[y][x]+=(double)z; SD[y][x]+=((double)z*(double)z);
	    }
	    fclose(hp);
	}

	c=(double)(argc-3);
	for (y=0; y<Ny; y++) for (x=0; x<Nx; x++) {
	    mean[y][x]/=c; SD[y][x]=sqrt(SD[y][x]/c-mean[y][x]*mean[y][x]);
	}

#ifdef	FUNCTION_CELL
	cell=mean; StoreImageFile(*(++argv),Nx,Ny,BPS,GetCell,NULL);
	cell=SD;   StoreImageFile(*(++argv),Nx,Ny,BPS,GetCell,NULL);
#else
{
	Cell	**cell;

	if ((cell=(Cell **)MA(Ny,cell))==NULL) Error(cmae);

	for (y=0; y<Ny; y++) {
	    if ((cell[y]=(Cell *)MA(Nx,*cell))==NULL) Error(cmae);

	    for (x=0; x<Nx; x++) cell[y][x]=ROUND(mean[y][x]);
	}
	StoreImageFile(*(++argv),Nx,Ny,BPS,cell,NULL);

	for (y=0; y<Ny; y++) for (x=0; x<Nx; x++) cell[y][x]=ROUND(SD[y][x]);

	StoreImageFile(*(++argv),Nx,Ny,BPS,cell,NULL);
}
#endif
	return 0;
}
