
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>

#define time	tick	/* to prevent the error of lint */

#define Error(msg)	{ fprintf(stderr,"%s\n",msg); exit(1); }
#define Abort(msg)	{ fprintf(stderr,"%s : %s\n",path,msg); exit(1); }
#define MA(cnt,ptr)	malloc((cnt)*sizeof(*(ptr)))

#define GetByte(file)	fgetc(file)

static int	GetWord(file)
FILE		*file;
{
	int	lo,hi;

	return ((lo=GetByte(file))==EOF ||
		(hi=GetByte(file))==EOF)?-1:(hi<<8)|lo;
}

#define DATE_LEN	256

static double	GetTime(path,len,file)
char		*path;
int		len;
FILE		*file;
{
	int		code,comma=0;
	char		date[DATE_LEN],*ptr=date;
	struct tm	tm;

	while (--len>=0)
	    if (GetByte(file)==',' && ++comma==11) break;

	if (comma!=11) Abort("top of imaging date not found.");

	while (--len>=0) {
	    if ((code=GetByte(file))==',')
		if (++comma==13)
		    break;
		else
		    code=' ';
	    else
		if (code=='-' || code==':') code=' ';

	    *(ptr++)=code;
	}

	if (comma!=13) Abort("end of imaging date not found.");

	*ptr='\0';

	if (sscanf(date,"%d %d %d %d %d %d",&(tm.tm_year),
					    &(tm.tm_mon ),
					    &(tm.tm_mday),
					    &(tm.tm_hour),
					    &(tm.tm_min ),
					    &(tm.tm_sec ))!=6)
	    Abort("bad format of imaging date.");

	while (--len>=0) (void)GetByte(file);

	tm.tm_year-=1900; tm.tm_mon-=1; return (double)mktime(&tm);
}

static int	Nx,Ny,y,firstImage=1;

static void	ReadImage(path,time,line)
char		*path;
double		*time,**line;
{
	FILE	*file=fopen(path,"rb");
	int	len,x,z;

	if (file==NULL) Abort("file not found.");

	if (GetByte(file)!='I' ||
	    GetByte(file)!='M') Abort("bad magic number.");

	if ((len=GetWord(file))<0) Abort("bad comment length.");

	if (firstImage) {
	    firstImage=0;

	    if ((Nx=GetWord(file))<=0 ||
		(Ny=GetWord(file))<=0) Abort("bad image size.");

	    if (y<0 || y>=Ny) Error("bad scanline position.");
	}
	else
	    if (GetWord(file)!=Nx ||
		GetWord(file)!=Ny) Abort("image size not match.");

	fseek(file,4L,1);	/* image offset */

	if (GetWord(file)!=2) Abort("bad image type.");

	fseek(file,50L,1);	/* reserved data */

	if (time==NULL)
	    fseek(file,(long)len,1);
	else
	    *time=GetTime(path,len,file);

	if ((*line=(double *)MA(Nx,*line))==NULL)
	    Abort("scanline memory allocation error.");

	fseek(file,((long)Nx*(long)y)*2L,1);

	for (x=0; x<Nx; x++) {
	    if ((z=GetWord(file))<0) Abort("unexpected end of file.");

	    (*line)[x]=(double)z;
	}
	fclose(file);
}

#define TEXT_LEN	512
#define PATH_LEN	256
#define COMMENT_CODE	'#'

static int	GetItem(path,name,time,file)
char		*path,*name;
double		*time;
FILE		*file;
{
	char	text[TEXT_LEN],*ptr;
	int	cnt=0;

	while (fgets(text,TEXT_LEN,file)!=NULL) {
	    if (text[strlen(text)-1]!='\n') Abort("too long line.");

	    if ((ptr=strchr(text,COMMENT_CODE))!=NULL) *ptr='\0';

	    if ((cnt=sscanf(text,"%s %lf",name,time))!=0) break;
	}
	return cnt;
}

static void	SortData(time,line,bgn,end)
double		*time,**line;
int		bgn,end;
{
	int	b=bgn,e=end;
	double	t,*l,key=time[(b+e)/2];

	do {
	    while (time[b]<key) ++b;
	    while (time[e]>key) --e;

	    if (b<=e) {
		if (b!=e) {
		    t=time[b]; time[b]=time[e]; time[e]=t;
		    l=line[b]; line[b]=line[e]; line[e]=l;
		}
		++b; --e;
	    }
	} while (b<=e);

	if (bgn<e) SortData(time,line,bgn,e);
	if (b<end) SortData(time,line,b,end);
}

static void	StoreMean(time,line,m0,m1,m2)
double		*time,**line;
int		m0,m1,m2;
{
	int	x,m,dm=m2-m1;

	if (m0!=m1) {
	    time[m0]=time[m1];
	    for (x=0; x<Nx; x++) line[m0][x]=line[m1][x];
	}
	for (m=m1+1; m<m2; m++) {
	    time[m0]+=time[m];
	    for (x=0; x<Nx; x++) line[m0][x]+=line[m][x];
	}
	if (dm!=1) {
	    time[m0]/=(double)dm;
	    for (x=0; x<Nx; x++) line[m0][x]/=(double)dm;
	}
}

#define TIME_UNIT	0.0001

static int	firstTime=1;
static double	time1,time2;

typedef struct {
		int	No;
		double	*time,**line;
	} Table;

static void	ReadTable(path,table,di)
char		*path;
Table		*table;
int		di;
{
	FILE	*file=fopen(path,"r");
	char	name[PATH_LEN];
	int	No,m,m1,m2,t1,t2;
	double	t,*time,**line;

	if (file==NULL) Abort("file not found.");

	switch(GetItem(path,name,&t,file)) {
	    case 0 :
		Abort("unexpected end of file.");
	    case 1 :
		if (sscanf(name,"%d",&No)==1) {
		    if (No<=0) Abort("bad number of lines.");

		    break;
		}
	    case 2 :
		for (No=1; GetItem(path,name,&t,file)!=0; No++) ;

		fseek(file,0L,0); break;
	}

	m=(di && No==1)?2:No;
	if ((time=(double  *)MA(m,time))==NULL ||
	    (line=(double **)MA(m,line))==NULL)
	    Abort("table memory allocation error.");

	for (m=0; m<No; m++) {
	    ReadImage(name,(GetItem(path,name,&t,file)==2)?NULL:&t,&(line[m]));

	    if (firstTime) {
		firstTime=0; time1=time2=t;
	    }
	    else {
		if (t<time1) time1=t;
		if (t>time2) time2=t;
	    }
	    time[m]=t;
	}
	fclose(file);

	if (di) {
	    SortData(time,line,0,No-1);

	    t=time[0]; m=m1=t1=0;
	    for (m2=1; m2<No; m2++) {
		t2=(int)floor((time[m2]-t)/TIME_UNIT);
		if (t1!=t2) {
		    StoreMean(time,line,m,m1,m2);
		    ++m; m1=m2; t1=t2;
		}
	    }
	    StoreMean(time,line,m,m1,m2);
	    table->No=m+1;
	}
	else
	    table->No=No;

	table->time=time; table->line=line;
}

static void	CorrectIntensity(d,ie,str)
Table		*d,*ie;
char		*str;
{
	int	m,k,n,No2=d->No-2;
	double	t0,*t1,r1,r2,*l0,*l1,*l2;

	for (m=0; m<ie->No; m++) {
	    t0=ie->time[m];
	    for (t1=(&(d->time[k=No2])); k>0 && t0<*t1; k--, t1--) ;
	    r1=(t0-*t1)/(t1[1]-*t1); r2=1.0-r1; l1=d->line[k]; l2=d->line[k+1];
	    for (l0=ie->line[m], n=0; n<Nx; n++, l0++)
		if ((*l0-=(r2*(*(l1++))+r1*(*(l2++))))<=0.0)
		    fprintf(stderr,
			    "bad %s intensity at (%d, %d) on %d-th image.\n",
			    str,n,y,m);
	}
}

static double	LnI0I(I0,I)
double		I0,I;
{
	return (I0<=0.0 || I<=0.0)?0.0:log(I0/I);
}

static void	WriteProjection(path,i,e)
char		*path;
Table		*i,*e;
{
	FILE	*file;
	int	m,k,n,No2=i->No-2;
	double	t0,*t1,r1,r2,*l0,*l1,*l2,p;

	if (*path=='-')
	    file=stdout;
	else
	    if ((file=fopen(path,"wb"))==NULL) Abort("file not stored.");

	fprintf(file,"%d\t%d\n",Nx,e->No);

	for (m=0; m<e->No; m++) {
	    t0=e->time[m];
	    for (t1=(&(i->time[k=No2])); k>0 && t0<*t1; k--, t1--) ;
	    r1=(t0-*t1)/(t1[1]-*t1); r2=1.0-r1; l1=i->line[k]; l2=i->line[k+1];
	    for (l0=e->line[m], n=0; n<Nx; n++, l0++) {
		p=LnI0I(r2*(*(l1++))+r1*(*(l2++)),*l0);
		fwrite(&p,sizeof(p),1,file);
	    }
	}
	if (file!=stdout) fclose(file);
}

#ifdef	SPLINE_INTERPOLATION

#define SI	SPLINE_INTERPOLATION

static void	WriteProjectionWithSI(path,i,e)
char		*path;
Table		*i,*e;
{
	int	m,n,k,No=i->No,No1=No-1,No2=No-2;
	double	**g,*c,*b,h1,h2,C,B,a,D,t0,*t1,h,z,*l0,*l1,*l2,*g1,*g2,v,I0,p,
		*t=i->time,**l=i->line;
	char	*msiae="memory for spline interpolation allocation error.";
	FILE	*file;

	if ((g=(double **)MA(No, g))==NULL ||
	    (c=(double  *)MA(No1,c))==NULL ||
	    (b=(double  *)MA(No1,b))==NULL) Error(msiae);

	for (m=0; m<No; m++)
	    if ((g[m]=(double *)MA(Nx,*g))==NULL) Error(msiae);

	h1=t[1]-t[0]; b[0]=SI; C=c[0]=2.0;
	for (n=0; n<Nx; n++) g[0][n]=(SI+2.0)*(l[1][n]-l[0][n])/h1;

	for (m=1; m<=No2; m++) {
	    h2=t[m+1]-t[m]; a=h2/(h1+h2); B=b[m]=1.0-a; c[m]=2.0-a*b[m-1]/C;
	    for (n=0; n<Nx; n++)
		g[m][n]=3.0*(a*(l[m  ][n]-l[m-1][n])/h1+
			     B*(l[m+1][n]-l[m  ][n])/h2)-a*g[m-1][n]/C;

	    C=c[m]; h1=h2;
	}
	B=b[No2]; D=2.0-SI*B/C;
	for (n=0; n<Nx; n++)
	    g[No1][n]=((SI+2.0)*(l[No1][n]-l[No2][n])/h1-SI*g[No2][n]/C)/D;

	for (m=No2; m>=0; m--) {
	    B=b[m]; C=c[m];
	    for (n=0; n<Nx; n++) g[m][n]=(g[m][n]-B*g[m+1][n])/C;
	}

	if (*path=='-')
	    file=stdout;
	else
	    if ((file=fopen(path,"wb"))==NULL) Abort("file not stored.");

	fprintf(file,"%d\t%d\n",Nx,e->No);

	for (m=0; m<e->No; m++) {
	    t0=e->time[m];
	    for (t1=(&(t[k=No2])); k>0 && t0<*t1; k--, t1--) ;
	    h=t1[1]-*t1; z=(t0-*t1)/h; l1=l[k]; l2=l[k+1]; g1=g[k]; g2=g[k+1];
	    for (l0=e->line[m], n=0; n<Nx; n++, l0++) {
		v=(*l2)-*l1;
		if ((I0=((((    *g1+*g2)*h-2.0*v)*z-
			   (2.0**g1+*g2)*h+3.0*v)*z+*g1*h)*z+*l1)<=0.0)
		    fprintf(stderr,"bad projection at (%d, %d).\n",n,m);

		p=LnI0I(I0,*l0);
		fwrite(&p,sizeof(p),1,file);
		++l1; ++l2; ++g1; ++g2;
	    }
	}
	if (file!=stdout) fclose(file);

	for (m=No1; m>=0; m--) free(g[m]); free(b); free(c); free(g);
}
#endif

int	main(argc,argv)
int	argc;
char	**argv;
{
	Table	dark,incident,emergent;

	if (argc!=6)
	    Error("usage : hp2p y dark.tbl incident.tbl emergent.tbl XP");

	y=atoi(argv[1]);

	ReadTable(argv[2],&dark,    1);
	ReadTable(argv[3],&incident,1);
	ReadTable(argv[4],&emergent,0);

	if (dark.No==1) {
	    dark.No=2;
	    dark.time[0]=time1*2.0-time2;
	    dark.time[1]=time2*2.0-time1;
	    dark.line[1]=dark.line[0];
	}

	CorrectIntensity(&dark,&incident,"incident");
	CorrectIntensity(&dark,&emergent,"emergent");

	if (incident.No==1) {
	    incident.No=2;
	    incident.time[0]=time1*2.0-time2;
	    incident.time[1]=time2*2.0-time1;
	    incident.line[1]=incident.line[0];
	}

#ifdef	SPLINE_INTERPOLATION
	if (incident.No>=3)
	    WriteProjectionWithSI(argv[5],&incident,&emergent);
	else
#endif
	WriteProjection(argv[5],&incident,&emergent);

	return 0;
}
