
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

static void	Abort(path,msg)
char		*path,*msg;
{
	fprintf(stderr,"%s : %s\n",path,msg); exit(1);
}

#define GetByte(file)	fgetc(file)

static int	GetWord(file)
FILE		*file;
{
	int	lo,hi;

	return ((lo=GetByte(file))==EOF ||
		(hi=GetByte(file))==EOF)?-1:(hi<<8)|lo;
}

static FILE	*OpenFile(path,Nx,Ny,firstFile)
char		*path;
int		*Nx,*Ny,firstFile;
{
	FILE	*file=fopen(path,"rb");
	int	len;

	if (file==NULL) Abort(path,"file not open.");

	if (GetByte(file)!='I' ||
	    GetByte(file)!='M') Abort(path,"bad magic number.");

	if ((len=GetWord(file))<0) Abort(path,"bad comment length.");

	if (firstFile) {
	    if ((*Nx=GetWord(file))<=0 ||
		(*Ny=GetWord(file))<=0) Abort(path,"bad image size.");
	}
	else
	    if (*Nx!=GetWord(file) ||
		*Ny!=GetWord(file)) Abort(path,"image size not match.");

	fseek(file,4L,1);		/* image offset */

	if (GetWord(file)!=2) Abort(path,"bad image type.");

	fseek(file,(long)(len+50),1);	/* reserved data and comment */

	return file;
}

static double	GetData(path,file)
char		*path;
FILE		*file;
{
	int	word;

	if ((word=GetWord(file))<0) Abort(path,"unexpected end of file.");

	return (double)word;
}

#define PathD	argv[1]
#define PathI	argv[2]
#define PathE	argv[3]

int	main(argc,argv)
int	argc;
char	**argv;
{
	FILE	*fileD,*fileI,*fileE;
	int	Nx,Ny,x,y;
	double	dataD,p,sum,sqr,ave;

	if (argc!=4) {
	    fputs("usage : hp2psv dark incident emergent\n",stderr); exit(1);
	}

	fileD=OpenFile(PathD,&Nx,&Ny,1);
	fileI=OpenFile(PathI,&Nx,&Ny,0);
	fileE=OpenFile(PathE,&Nx,&Ny,0);

	printf("%d\t%d\n",Nx,Ny);

	for (y=0; y<Ny; y++) {
	    sum=sqr=0.0;
	    for (x=0; x<Nx; x++) {
		dataD=GetData(PathD,fileD);
		p=log((GetData(PathI,fileI)-dataD)/
		      (GetData(PathE,fileE)-dataD));
		sum+=p; sqr+=(p*p);
	    }
	    ave=sum/(double)Nx;
	    printf("%d\t%lf\t%lf\n",y,sum,sqr/(double)Nx-ave*ave);
	}
	return 0;
}
