
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>

#define time	tick	/* to prevent the error of lint */

#define Error(msg)	{ fprintf(stderr,"%s\n",msg); exit(1); }
#define Abort(msg)	{ fprintf(stderr,"%s : %s\n",path,msg); exit(1); }
#define MA(cnt,ptr)	malloc((cnt)*sizeof(*(ptr)))

#define GetByte(file)	fgetc(file)

static int	GetWord(file)
FILE		*file;
{
	int	lo,hi;

	return ((lo=GetByte(file))==EOF ||
		(hi=GetByte(file))==EOF)?-1:(hi<<8)|lo;
}

#define DATE_LEN	256

static double	GetTime(path,len,file)
char		*path;
int		len;
FILE		*file;
{
	int		code,comma=0;
	char		date[DATE_LEN],*ptr=date;
	struct tm	tm;

	while (--len>=0)
	    if (GetByte(file)==',' && ++comma==11) break;

	if (comma!=11) Abort("top of imaging date not found.");

	while (--len>=0) {
	    if ((code=GetByte(file))==',')
		if (++comma==13)
		    break;
		else
		    code=' ';
	    else
		if (code=='-' || code==':') code=' ';

	    *(ptr++)=code;
	}

	if (comma!=13) Abort("end of imaging date not found.");

	*ptr='\0';

	if (sscanf(date,"%d %d %d %d %d %d",&(tm.tm_year),
					    &(tm.tm_mon ),
					    &(tm.tm_mday),
					    &(tm.tm_hour),
					    &(tm.tm_min ),
					    &(tm.tm_sec ))!=6)
	    Abort("bad format of imaging date.");

	while (--len>=0) (void)GetByte(file);

	tm.tm_year-=1900; tm.tm_mon-=1; return (double)mktime(&tm);
}

static int	Nx,Ny,Nxy,firstImage=1;

static void	ReadImage(path,time,line)
char		*path;
double		*time,**line;
{
	FILE	*file=fopen(path,"rb");
	int	len,n,z;

	if (file==NULL) Abort("file not found.");

	if (GetByte(file)!='I' ||
	    GetByte(file)!='M') Abort("bad magic number.");

	if ((len=GetWord(file))<0) Abort("bad comment length.");

	if (firstImage) {
	    firstImage=0;

	    if ((Nx=GetWord(file))<=0 ||
		(Ny=GetWord(file))<=0) Abort("bad image size.");

	    Nxy=Nx*Ny;
	}
	else
	    if (GetWord(file)!=Nx ||
		GetWord(file)!=Ny) Abort("image size not match.");

	fseek(file,4L,1);	/* image offset */

	if (GetWord(file)!=2) Abort("bad image type.");

	fseek(file,50L,1);	/* reserved data */

	if (time==NULL)
	    fseek(file,(long)len,1);
	else
	    *time=GetTime(path,len,file);

	if (line!=NULL) {
	    if ((*line=(double *)MA(Nxy,*line))==NULL)
		Abort("scanline memory allocation error.");

	    for (n=0; n<Nxy; n++) {
		if ((z=GetWord(file))<0) Abort("unexpected end of file.");

		(*line)[n]=(double)z;
	    }
	}
	fclose(file);
}

#define TEXT_LEN	512
#define PATH_LEN	256
#define COMMENT_CODE	'#'

static int	GetItem(path,name,time,file)
char		*path,*name;
double		*time;
FILE		*file;
{
	char	text[TEXT_LEN],*ptr;
	int	cnt=0;

	while (fgets(text,TEXT_LEN,file)!=NULL) {
	    if (text[strlen(text)-1]!='\n') Abort("too long line.");

	    if ((ptr=strchr(text,COMMENT_CODE))!=NULL) *ptr='\0';

	    if ((cnt=sscanf(text,"%s %lf",name,time))!=0) break;
	}
	return cnt;
}

static void	SortData(time,line,bgn,end)
double		*time,**line;
int		bgn,end;
{
	int	b=bgn,e=end;
	double	t,*l,key=time[(b+e)/2];

	do {
	    while (time[b]<key) ++b;
	    while (time[e]>key) --e;

	    if (b<=e) {
		if (b!=e) {
		    t=time[b]; time[b]=time[e]; time[e]=t;
		    l=line[b]; line[b]=line[e]; line[e]=l;
		}
		++b; --e;
	    }
	} while (b<=e);

	if (bgn<e) SortData(time,line,bgn,e);
	if (b<end) SortData(time,line,b,end);
}

static void	StoreMean(time,line,m0,m1,m2)
double		*time,**line;
int		m0,m1,m2;
{
	int	n,m,dm=m2-m1;

	if (m0!=m1) {
	    time[m0]=time[m1];
	    for (n=0; n<Nxy; n++) line[m0][n]=line[m1][n];
	}
	for (m=m1+1; m<m2; m++) {
	    time[m0]+=time[m];
	    for (n=0; n<Nxy; n++) line[m0][n]+=line[m][n];
	}
	if (dm!=1) {
	    time[m0]/=(double)dm;
	    for (n=0; n<Nxy; n++) line[m0][n]/=(double)dm;
	}
}

#define TIME_UNIT	0.0001

static int	firstTime=1;
static double	time1,time2;

typedef struct {
		int	No;
		double	*time;
		union {
			double	**line;
			char	**name;
		} lorn;
	} Table;

#define LINE	lorn.line
#define NAME	lorn.name

static void	ReadTable(path,table,di)
char		*path;
Table		*table;
int		di;
{
	FILE	*file=fopen(path,"r");
	char	name[PATH_LEN],*tmae="table memory allocation error.";
	int	No,m,m1,m2,t1,t2;
	double	t,*time,*tp;

	if (file==NULL) Abort("file not found.");

	switch(GetItem(path,name,&t,file)) {
	    case 0 :
		Abort("unexpected end of file.");
	    case 1 :
		if (sscanf(name,"%d",&No)==1) {
		    if (No<=0) Abort("bad number of lines.");

		    break;
		}
	    case 2 :
		for (No=1; GetItem(path,name,&t,file)!=0; No++) ;

		fseek(file,0L,0); break;
	}

	if (di) {
	    m=(No==1)?2:No;
	    if ((time=(double *)MA(m,time))==NULL ||
		(table->LINE=(double **)MA(m,table->LINE))==NULL) Abort(tmae);
	}
	else
	    if ((time=(double *)MA(No,time))==NULL ||
		(table->NAME=(char **)MA(No,table->NAME))==NULL) Abort(tmae);

	for (m=0; m<No; m++) {
	    tp=(GetItem(path,name,&t,file)==2)?NULL:&t;

	    if (di)
		ReadImage(name,tp,&(table->LINE[m]));
	    else {
		ReadImage(name,tp,NULL);

		if ((table->NAME[m]=(char *)malloc(strlen(name)+1))==NULL) {
		    fprintf(stderr,"%s : path name memory allocation error.\n",
			    name);
		    exit(1);
		}
		strcpy(table->NAME[m],name);
	    }

	    if (firstTime) {
		firstTime=0; time1=time2=t;
	    }
	    else {
		if (t<time1) time1=t;
		if (t>time2) time2=t;
	    }
	    time[m]=t;
	}
	fclose(file);

	if (di) {
	    SortData(time,table->LINE,0,No-1);

	    t=time[0]; m=m1=t1=0;
	    for (m2=1; m2<No; m2++) {
		t2=(int)floor((time[m2]-t)/TIME_UNIT);
		if (t1!=t2) {
		    StoreMean(time,table->LINE,m,m1,m2);
		    ++m; m1=m2; t1=t2;
		}
	    }
	    StoreMean(time,table->LINE,m,m1,m2);
	    table->No=m+1;
	}
	else
	    table->No=No;

	table->time=time;
}

static void	CorrectIntensity(d,ieNo,ieTime,ieLine,str)
Table		*d;
int		ieNo;
double		*ieTime,**ieLine;
char		*str;
{
	int	m,k,n,No2=d->No-2;
	double	t0,*t1,r1,r2,*l0,*l1,*l2;

	for (m=0; m<ieNo; m++) {
	    t0=ieTime[m];
	    for (t1=(&(d->time[k=No2])); k>0 && t0<*t1; k--, t1--) ;
	    r1=(t0-*t1)/(t1[1]-*t1); r2=1.0-r1; l1=d->LINE[k]; l2=d->LINE[k+1];
	    for (l0=ieLine[m], n=0; n<Nxy; n++, l0++)
		if ((*l0-=(r2*(*(l1++))+r1*(*(l2++))))<=0.0)
		    fprintf(stderr,
			    "bad %s intensity at (%d, %d) on %d-th image.\n",
			    str,n%Nx,n/Nx,m);
	}
}

static double	LnI0I(I0,I)
double		I0,I;
{
	return (I0<=0.0 || I<=0.0)?0.0:log(I0/I);
}

static void	CalculateProjection(i,t0,l0)
Table		*i;
double		t0,*l0;
{
	int	k,n,No2=i->No-2;
	double	*t1,r1,r2,*l1,*l2;

	for (t1=(&(i->time[k=No2])); k>0 && t0<*t1; k--, t1--) ;
	r1=(t0-*t1)/(t1[1]-*t1); r2=1.0-r1; l1=i->LINE[k]; l2=i->LINE[k+1];
	for (n=0; n<Nxy; n++) {
	    *l0=LnI0I(r2*(*l1)+r1*(*l2),*l0); ++l0; ++l1; ++l2;
	}
}

#ifdef	SPLINE_INTERPOLATION

#define SI	SPLINE_INTERPOLATION

static double	**SetUpSI(i)
Table		*i;
{
	int	m,n,No=i->No,No1=No-1,No2=No-2;
	double	**g,*c,*b,h1,h2,C,B,a,D,*t=i->time,**l=i->LINE;
	char	*nmsi="no memory for spline interpolation (warning).\n";

	if ((g=(double **)MA(No,g))==NULL) {
	    fputs(nmsi,stderr); return NULL;
	}
	for (m=0; m<No; m++)
	    if ((g[m]=(double *)MA(Nxy,*g))==NULL) {
		while (--m>=0) free(g[m]); free(g);
		fputs(nmsi,stderr); return NULL;
	    }

	if ((c=(double *)MA(No1,c))==NULL) {
	    for (m=No-1; m>=0; m--) free(g[m]); free(g);
	    fputs(nmsi,stderr); return NULL;
	}
	if ((b=(double *)MA(No1,b))==NULL) {
	    free(c); for (m=No-1; m>=0; m--) free(g[m]); free(g);
	    fputs(nmsi,stderr); return NULL;
	}

	h1=t[1]-t[0]; b[0]=SI; C=c[0]=2.0;
	for (n=0; n<Nxy; n++) g[0][n]=(SI+2.0)*(l[1][n]-l[0][n])/h1;

	for (m=1; m<=No2; m++) {
	    h2=t[m+1]-t[m]; a=h2/(h1+h2); B=b[m]=1.0-a; c[m]=2.0-a*b[m-1]/C;
	    for (n=0; n<Nxy; n++)
		g[m][n]=3.0*(a*(l[m  ][n]-l[m-1][n])/h1+
			     B*(l[m+1][n]-l[m  ][n])/h2)-a*g[m-1][n]/C;

	    C=c[m]; h1=h2;
	}
	B=b[No2]; D=2.0-SI*B/C;
	for (n=0; n<Nxy; n++)
	    g[No1][n]=((SI+2.0)*(l[No1][n]-l[No2][n])/h1-SI*g[No2][n]/C)/D;

	for (m=No2; m>=0; m--) {
	    B=b[m]; C=c[m];
	    for (n=0; n<Nxy; n++) g[m][n]=(g[m][n]-B*g[m+1][n])/C;
	}
	free(b); free(c); return g;
}

static void	CalculateProjectionWithSI(i,g,t0,l0)
Table		*i;
double		**g,t0,*l0;
{
	int	k,n;
	double	*t1,h,z,*l1,*l2,*g1,*g2,v,I0,*t=i->time,**l=i->LINE;

	for (t1=(&(t[k=i->No-2])); k>0 && t0<*t1; k--, t1--) ;
	h=t1[1]-*t1; z=(t0-*t1)/h; l1=l[k]; l2=l[k+1]; g1=g[k]; g2=g[k+1];
	for (n=0; n<Nxy; n++) {
	    v=(*l2)-*l1;
	    if ((I0=((((    *g1+*g2)*h-2.0*v)*z-
		       (2.0**g1+*g2)*h+3.0*v)*z+*g1*h)*z+*l1)<=0.0)
		fprintf(stderr,"bad projection at (%d, %d).\n",n%Nx,n/Nx);

	    *l0=LnI0I(I0,*l0); ++l0; ++l1; ++l2; ++g1; ++g2;
	}
}
#endif

int	main(argc,argv)
int	argc;
char	**argv;
{
	Table	d,i,e;
	double	**ta,*line,sum,dr=1.0;
	char	*amae="array memory allocation error.";
	int	y,m,n,x;
	FILE	*file;

#ifdef	SPLINE_INTERPOLATION
	double	**g=NULL;
#endif

	if (argc!=5 && argc!=6)
	    Error("usage : hp2ta d.tbl i.tbl e.tbl {dr} XTA");

	ReadTable(argv[1],&d,1);
	ReadTable(argv[2],&i,1);
	ReadTable(argv[3],&e,0);

	if ((ta=(double **)MA(Ny,*ta))==NULL) Error(amae);

	for (y=0; y<Ny; y++) {
	    if ((ta[y]=(double *)MA(e.No,*ta))==NULL) Error(amae);

	    for (m=0; m<e.No; m++) ta[y][m]=0.0;
	}
	if (argc==6) dr=atof(argv[4]);

	if (d.No==1) {
	    d.No=2; d.time[0]=time1*2.0-time2;
		    d.time[1]=time2*2.0-time1; d.LINE[1]=d.LINE[0];
	}

	CorrectIntensity(&d,i.No,i.time,i.LINE,"incident");

	if (i.No==1) {
	    i.No=2; i.time[0]=time1*2.0-time2;
		    i.time[1]=time2*2.0-time1; i.LINE[1]=i.LINE[0];
	}

#ifdef	SPLINE_INTERPOLATION
	if (i.No>=3) g=SetUpSI(&i);
#endif

	for (m=0; m<e.No; m++) {
	    ReadImage(e.NAME[m],NULL,&line);

	    CorrectIntensity(&d,1,e.time,&line,"emergent");

#ifdef	SPLINE_INTERPOLATION
	    if (i.No>=3 && g!=NULL)
		CalculateProjectionWithSI(&i,g,e.time[m],line);
	    else
#endif
	    CalculateProjection(&i,e.time[m],line);

	    for (n=0, y=0; y<Ny; y++) {
		sum=0.0;
		for (x=0; x<Nx; x++, n++)
#ifdef	PP
		    if (line[n]>0.0)
#endif
		    sum+=line[n];

		ta[y][m]=sum*dr;
	    }
	    free(line);
	}

	if (argv[argc-1][0]=='-')
	    file=stdout;
	else
	    if ((file=fopen(argv[argc-1],"wb"))==NULL) {
		fprintf(stderr,"%s : file not stored.\n",argv[argc-1]); exit(1);
	    }

	fprintf(file,"%d\t%d\n",e.No,Ny);

	for (y=0; y<Ny; y++)
#ifdef	NORMALIZE
{
	double	mean,nta,sum1=0.0,sum2=0.0;

	for (m=0; m<e.No; m++) {
	    sum1+=ta[y][m]; sum2+=(ta[y][m]*ta[y][m]);
	}
	mean=sum1/(double)e.No;
	fprintf((file!=stdout)?stdout:stderr,
		"%d\t%lf\t%lf\n",y,mean,sqrt(sum2/(double)e.No-mean*mean));

	for (m=0; m<e.No; m++) {
	    nta=ta[y][m]/mean; fwrite(&nta,sizeof(nta),1,file);
	}
}
#else
	    for (m=0; m<e.No; m++)
		fwrite(&(ta[y][m]),sizeof(ta[y][m]),1,file);
#endif
	if (file!=stdout) fclose(file); return 0;
}
