
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "sif.h"

#ifdef	TXY
#define log(I0I)	(1.0/(I0I))
#endif

static void	Abort(name,msg)
char		*name,*msg;
{
	fprintf(stderr,"%s : %s.\n",name,msg); exit(1);
}

static int	FGetW(file)
FILE		*file;
{
	int	lo,hi;

	return ((lo=fgetc(file))==EOF ||
		(hi=fgetc(file))==EOF)?-1:(hi<<8)|lo;
}

static FILE	*Open(name,check,Wx,Wy)
char		*name;
int		check,*Wx,*Wy;
{
	FILE	*file=fopen(name,"rb");
	int	len,idx;

	if (file==NULL) Abort(name,"file not found");

	if (fgetc(file)!='I' ||
	    fgetc(file)!='M') Abort(name,"bad magic number");

	if ((len=FGetW(file))<0) Abort(name,"bad comment length");

	if (check) {
	    if (*Wx!=FGetW(file) ||
		*Wy!=FGetW(file)) Abort(name,"image size not match");
	}
	else
	    if ((*Wx=FGetW(file))<=0 ||
		(*Wy=FGetW(file))<=0) Abort(name,"bad image size");

	for (idx=4; idx>0; idx--) (void)fgetc(file);

	if (FGetW(file)!=2) Abort(name,"bad image type");

	for (idx=50+len; idx>0; idx--) (void)fgetc(file);

	return file;
}

static int	FGetI(name,file)
char		*name;
FILE		*file;
{
	int	I;

	if ((I=FGetW(file))<0) Abort(name,"unexpected end of file");

	return I;
}

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

#define STR_LEN	256
#define NO_DATA	0

int	main(argc,argv)
int	argc;
char	**argv;
{
	FILE	*Fd,*Fi,*Ft;
	int	Wx,Wy,y,x,Id,Ii,It,BPS;
	double	**p,p1,p2,p0,dp;
	Cell	lim,**cell;
	char	str[STR_LEN],*mae="memory allocation error.";

	if (argc!=6 && argc!=8)
#ifdef	TXY
	    Error("usage : hp3txy d.img i.img t.img {t0 dt} BPS t.tif");
#else
	    Error("usage : hp3pxy d.img i.img t.img {p0 dp} BPS p.tif");
#endif

#define Nd	argv[1]
#define Ni	argv[2]
#define Nt	argv[3]

	Fd=Open(Nd,0,&Wx,&Wy);
	Fi=Open(Ni,1,&Wx,&Wy);
	Ft=Open(Nt,1,&Wx,&Wy);

	if ((p=(double **)malloc(Wy*sizeof(*p)))==NULL) Error(mae);

	p1=HUGE_VAL; p2=0.0;
	for (y=0; y<Wy; y++) {
	    if ((p[y]=(double *)malloc(Wx*sizeof(**p)))==NULL) Error(mae);

	    for (x=0; x<Wx; x++) {
		Id=FGetI(Nd,Fd);
		Ii=FGetI(Ni,Fi);
		It=FGetI(Nt,Ft);
		if (Ii>Id && It>Id) {
		    p[y][x]=log((double)(Ii-Id)/(double)(It-Id));
		    if (p1>p2)
			p1=p2=p[y][x];
		    else {
			if (p[y][x]<p1) p1=p[y][x];
			if (p[y][x]>p2) p2=p[y][x];
		    }
		}
		else
		    p[y][x]=HUGE_VAL;
	    }
	}
	if (p1>=p2) Error("no meaningful intensity data.");

	printf("%lf\t%lf\n",p1,p2);

	if ((BPS=atoi(argv[argc-2]))<=0 || BPS>8*sizeof(Cell))
	    Error("bad bits per sample (BPS).");

	lim=(Cell)((1<<BPS)-1);

	if (argc==6) {
	    sprintf(str,"%lf\t%lf",p1,p2);

	    p0=p1;
	    dp=(p2-p1)/(double)lim;
	}
	else {
	    sprintf(str,"%lf\t%lf\n%s\t%s",p1,p2,argv[4],argv[5]);

	    if ((dp=atof(argv[5]))<=0.0) Error("bad value step.");

	    p1=p0=atof(argv[4]);
	    p2=p0+dp*(double)lim;
	}

	cell=(Cell **)p;
	for (y=0; y<Wy; y++)
	for (x=0; x<Wx; x++)
	    if (p[y][x]!=HUGE_VAL)
		if (p[y][x]<p1)
		    cell[y][x]=0;
		else if (p[y][x]>p2)
		    cell[y][x]=lim;
		else
		    cell[y][x]=(Cell)floor((p[y][x]-p0)/dp+0.5);
	    else
		cell[y][x]=NO_DATA;

	StoreImageFile(argv[argc-1],Wx,Wy,BPS,cell,str);

	return 0;
}
