
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"
#include "pos.h"

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

typedef struct Corner {
		int		x,y;
		struct Corner	*next;
	} Corner;

static struct {
		int	dx,dy;
	} base[4]={{0,0},{0,0},{-1,0},{0,-1}},
	  enws[4]={{1,0},{0,1},{-1,0},{0,-1}};

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nx,Ny,BPS,background,threshold,thickness,
		x,y,num,dir1,x1,y1,step,dir2,x2,y2,xMin,yMin,xMax,yMax,tot=0;
	Cell	**cell;
	char	**flag,**edge[2];
	Corner	*top,**ptr,*next;

	if (argc!=3 && argc!=4) Error("usage : i2e TIFF threshold {thickness}");

	ReadImageFile(argv[1],&Nx,&Ny,&BPS,&cell,NULL);
	fprintf(stderr,"%d\t%d\n",Nx,Ny);

	if ((threshold=atoi(argv[2]))<=0 || threshold>=(1<<BPS))
	    Error("bad threshold level.");

	if (argc==3)
	    background=thickness=0;
	else
	    if (argv[3][0]=='-') {
		background=0; thickness=atoi(&(argv[3][1]));
	    }
	    else {
		background=1; thickness=atoi(argv[3]);
	    }

	if (thickness<0) Error("bad object skin thickness.");

	flag=PareObjectSkin(Nx,Ny,cell,background,threshold,thickness);

	if ((edge[0]=AllocateFlagMemory(Nx,Ny+1))==NULL ||
	    (edge[1]=AllocateFlagMemory(Nx+1,Ny))==NULL)
	    Error("edge memory allocation error.");

	for (x=0; x<Nx; x++) {
	    edge[0][0][x]=flag[0][x];
	    for (y=1; y<Ny; y++) edge[0][y][x]=flag[y-1][x]^flag[y][x];
	    edge[0][Ny][x]=flag[Ny-1][x];
	}
	for (y=0; y<Ny; y++) {
	    edge[1][y][0]=flag[y][0];
	    for (x=1; x<Nx; x++) edge[1][y][x]=flag[y][x-1]^flag[y][x];
	    edge[1][y][Nx]=flag[y][Nx-1];
	}

	xMin=Nx; yMin=Ny; xMax=yMax=0;

	for (y=0; y<Ny; y++) for (x=0; x<Nx; x++) if (edge[0][y][x]) {
	    top=NULL; ptr=(&top); num=0; dir1=(flag[y][x])?3:0; x1=x; y1=y;
	    for (step=0; step<3; step++) {
		dir2=(dir1-step+5)%4;
		if ((x2=x1+base[dir2].dx)<0 ||
		    (y2=y1+base[dir2].dy)<0) continue;

		switch(dir2) {
		    case 0 : case 2 :
			if (x2>=Nx || y2>Ny || !edge[0][y2][x2]) continue;
			edge[0][y2][x2]=0; break;
		    case 1 : case 3 :
			if (x2>Nx || y2>=Ny || !edge[1][y2][x2]) continue;
			edge[1][y2][x2]=0; break;
		}
		if (dir1!=dir2) {
		    if ((*ptr=(Corner *)malloc(sizeof(**ptr)))==NULL)
			Error("corner memory allocation error.");

		    (*ptr)->x=x1; (*ptr)->y=y1; (*ptr)->next=NULL;
		    ptr=(&((*ptr)->next)); ++num; dir1=dir2;
		}
		x1+=enws[dir1].dx;
		y1+=enws[dir1].dy; step=(x1==x && y1==y)?2:(-1);
	    }
	    tot+=num;

	    printf("%d\n",num);
	    while (top!=NULL) {
		if (top->x<xMin) xMin=top->x; if (top->x>xMax) xMax=top->x;
		if (top->y<yMin) yMin=top->y; if (top->y>yMax) yMax=top->y;

		printf("%d\t%d\n",top->x,top->y);
		next=top->next; free(top); top=next;
	    }
	}
	if (tot==0) Error("no edge exists.");

	fprintf(stderr,"%d\n%d\t%d\t%d\t%d\n",tot,xMin,yMin,xMax,yMax);
	return 0;
}
