
#include <stdio.h>
#include <stdlib.h>
#include "CBP.h"
#include "sri.h"

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

#define DESC_LEN	256
#define RADIAN		0.01745329251994329576

int	main(argc,argv)
int	argc;
char	**argv;
{
	FILE	*file;
	int	N,M,n,m,bps,v,h;
	double	dr,r0,t0,p,f0,df,f1,f2;
	CBP	*cbp;
	Float	**f;
	char	desc[DESC_LEN];

	if (argc!=4 && argc!=6) Error("usage : p2i XP {f0 df} BPS TIFF");

	if (argv[1][0]=='-')
	    file=stdin;
	else
	    if ((file=fopen(argv[1],"r"))==NULL)
		Error("X-ray projection data file not found.");

	fscanf(file,"%d %lf %lf %d %lf",&N,&dr,&r0,&M,&t0);
	if (N<=0 || dr<=0.0 || M<=0) Error("bad X-ray CT scanner parameter.");

	if ((cbp=SetUpCBP(N,M,dr,r0,t0*RADIAN))==NULL)
	    Error("SetUpCBP() memory allocation error.");

	for (m=0; m<M; m++) for (n=0; n<N; n++) {
	    fscanf(file,"%lf",&p); cbp->pq[m][n]=p;
	}
	if (file!=stdin) fclose(file);

	if (argc==6) {
	    f0=atof(argv[2]);
	    if ((df=atof(argv[3]))<=0.0) Error("bad CT value step.");
	}
	if ((bps=atoi(argv[argc-2]))<=0) Error("bad reconstruct image BPS.");

	Convolution(cbp);

	if (!ReGridingCP(cbp)) Error("ReGridingCP() memory allocation error.");

	if ((f=BPImage(cbp))==NULL) Error("BPImage() memory allocation error.");

	f1=f2=f[0][0];
	for (v=0; v<N; v++)
	for (h=0; h<N; h++) if (f[v][h]<f1) f1=f[v][h]; else
			    if (f[v][h]>f2) f2=f[v][h];

	fprintf(stderr,"%lf\t%lf\n",f1,f2);

	if (argc==6)
	    sprintf(desc,"%lf\t%lf\n%d\t%lf\n%lf\t%lf\n%lf\t%lf",
			 dr,r0,M,t0,f1,f2,f0,df);
	else {
	    f0=f1;
	    df=(f2-f1)/(double)((1<<bps)-1);

	    sprintf(desc,"%lf\t%lf\n%d\t%lf\n%lf\t%lf",dr,r0,M,t0,f1,f2);
	}
	StoreReconstructImage(argv[argc-1],N,f,f0-df/2.0,df,bps,desc);

	return 0;
}
