
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "cell.h"
#include "rif.h"
#include "sif.h"

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

#define STR_LEN	256

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	x1,y1,x2,y2,x0,y0,Nx,Ny,**cnt,x,y,BPS;
	char	path[STR_LEN],desc[STR_LEN];
	double	base,step,**sum,bg,val,min,max;

	struct List {
		char		*path;
		double		base,step;
		int		x0,y0,Nx,Ny;
		struct List	*next;
	} *list,*top=NULL,**ptr=(&top);

	Cell	**cell;

	if (argc!=4 && argc!=6 && argc!=8 && argc!=10)
	    Error("usage : ptvi {x1 y1 x2 y2} bg {base step} BPS TIFF");

	if (argc==8 || argc==10)
	    if ((x1=atoi(argv[1]))>(x2=atoi(argv[3])) ||
		(y1=atoi(argv[2]))>(y2=atoi(argv[4]))) Error("bad area.");

	while (scanf("%s %lf %lf %d %d",path,&base,&step,&x0,&y0)==5) {
	    ReadImageFile(path,&Nx,&Ny,NULL,NULL,NULL);

	    if (argc!=8 && argc!=10 && top==NULL) {
		x1=x0; x2=x0+Nx-1;
		y1=y0; y2=y0+Ny-1;
	    }
	    else {
		if (x0<x1) x1=x0; if ((x=x0+Nx-1)>x2) x2=x;
		if (y0<y1) y1=y0; if ((y=y0+Ny-1)>y2) y2=y;
	    }

	    if ((list=(struct List *)malloc(sizeof(*list)))==NULL ||
		(list->path=(char *)malloc(strlen(path)+1))==NULL)
		Error("parameter list allocation error.");

	    strcpy(list->path,path);
	    list->base=base; list->step=step;
	    list->x0=x0; list->y0=y0; list->Nx=Nx; list->Ny=Ny;
	    list->next=NULL; *ptr=list; ptr=(&(list->next));
	}
	if (top==NULL) Error("no TIFF value image.");

	x0=x1; Nx=x2-x1+1;
	y0=y1; Ny=y2-y1+1;

	if ((cnt=(int **)malloc(Ny*sizeof(*cnt)))==NULL ||
	    (sum=(double **)malloc(Ny*sizeof(*sum)))==NULL)
	    Error("array memory allocation error (1).");

	for (y=0; y<Ny; y++) {
	    if ((cnt[y]=(int *)malloc(Nx*sizeof(**cnt)))==NULL ||
		(sum[y]=(double *)malloc(Nx*sizeof(**sum)))==NULL)
		Error("array memory allocation error (2).");

	    for (x=0; x<Nx; x++) {
		cnt[y][x]=0; sum[y][x]=0.0;
	    }
	}

	for (list=top; list!=NULL; list=list->next) {
	    ReadImageFile(list->path,&x,&y,NULL,&cell,NULL);
	    if (list->Nx!=x || list->Ny!=y) {
		fprintf(stderr,"%s : ",list->path); Error("bad image size.");
	    }

	    x1=list->x0-x0;
	    y1=list->y0-y0;
	    base=list->base;
	    step=list->step;

	    for (y2=y1, y=0; y<list->Ny; y++, y2++) {
		for (x2=x1, x=0; x<list->Nx; x++, x2++) {
		    ++(cnt[y2][x2]);
		    sum[y2][x2]+=(base+step*(double)cell[y][x]);
		}
		free(cell[y]);
	    }
	    free(cell);
	}

	bg=atof(argv[(argc==4 || argc==6)?1:5]);

	for (y=0; y<Ny; y++)
	for (x=0; x<Nx; x++) {
	    val=sum[y][x]=(cnt[y][x]>0)?(sum[y][x]/(double)cnt[y][x]):bg;
	    if (x==0 && y==0)
		min=max=val;
	    else
		if (val<min) min=val; else if (val>max) max=val;
	}

	if ((BPS=atoi(argv[argc-2]))<=0 || BPS>8*sizeof(Cell))
	    Error("bad bits per sample.");

	if (argc==4 || argc==8) {
	    base=min;
	    if (min==max) {
		fputs("bad value range (warning).\n",stderr);

		step=1.0;
	    }
	    else
		step=(max-min)/(double)((1<<BPS)-1);

	    sprintf(desc,"%lf\t%lf",min,max);
	}
	else {
	    base=atof(argv[(argc==6)?2:6]);
	    if ((step=atof(argv[(argc==6)?3:7]))==0.0) Error("bad value step.");

	    sprintf(desc,"%lf\t%lf\n%lf\t%lf",min,max,base,step);

	    min=base;
	    max=base+step*(double)((1<<BPS)-1);
	}
	puts(desc);

	cell=(Cell **)sum;
	for (y=0; y<Ny; y++) {
	    cell[y]=(Cell *)sum[y];
	    for (x=0; x<Nx; x++) {
		if ((val=sum[y][x])<min) val=min; else if (val>max) val=max;

		cell[y][x]=(Cell)floor((val-base)/step+0.5);
	    }
	}
	StoreImageFile(argv[argc-1],Nx,Ny,BPS,cell,desc);

	return 0;
}
