
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "sif.h"

static void	Abort(path,msg)
char		*path,*msg;
{
	fprintf(stderr,"%s : %s\n",path,msg); exit(1);
}

#define MA(cnt,ptr)	malloc((size_t)(cnt)*sizeof(*(ptr)))

#define LINE_LEN	512

static void	ReadProjection(path,Nx,Ny,p,firstFile)
char		*path;
int		*Nx,*Ny,firstFile;
double		***p;
{
	FILE	*file;
	char	line[LINE_LEN],*pmae="projection memory allocation error.";
	int	x,y;

	if (*path=='-')
	    file=stdin;
	else
	    if ((file=fopen(path,"rb"))==NULL) Abort(path,"file not open.");

	if (fgets(line,LINE_LEN,file)==NULL ||
	    sscanf(line,"%d %d",&x,&y)!=2) Abort(path,"bad image size.");

	if (firstFile) {
	    *Nx=x; *Ny=y;
	}
	else
	    if (*Nx!=x || *Ny!=y) Abort(path,"image size not match.");

	if ((*p=(double **)MA(*Ny,*p))==NULL) Abort(path,pmae);

	for (y=0; y<*Ny; y++) {
	    if (((*p)[y]=(double *)MA(*Nx,**p))==NULL) Abort(path,pmae);

	    for (x=0; x<*Nx; x++)
		if (fread(&((*p)[y][x]),sizeof(***p),1,file)!=1)
		    Abort(path,"unexpected end of file.");
	}
	if (file!=stdin) fclose(file);
}

static double	Sqrt(d)
double		d;
{
	return (d<0.0)?0.0:sqrt(d);
}

static double	CalculateRMS(N1,dx,x1,x2,p000,p180)
int		N1,dx,x1,x2;
double		*p000,*p180;
{
	int	x;
	double	d,sum=0.0;

	for (x=x1; x<x2; x++) {
	    d=p000[x+dx]-p180[N1-x]; sum+=(d*d);
	}
	return Sqrt(sum/(double)(x2-x1));
}

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

#define BPS	(8*sizeof(Cell))

#define DESC_LEN	256

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nx,Ny,z1,z2,N1,N2,Nc,y,dx0,x1,x2,dx,x;
	double	**p000,**p180,**array,RMS0,RMS,R0,CC,D,A,B,RMS1,RMS2,dRMS,
		sCC=0.0,sCCy=0.0,sCCy2=0.0,sCCR0=0.0,sCCR02=0.0,sCCyR0=0.0;
	char	desc[DESC_LEN],
		*cmae="cell memory allocation error.";
	Cell	**cell;

	if (argc<3 || argc>6) Error("usage : pxy2DO P000 P180 {y1 y2} {TIFF}");

	ReadProjection(argv[1],&Nx,&Ny,&p000,1);
	ReadProjection(argv[2],&Nx,&Ny,&p180,0);

	if (argc==3 || argc==4) {
	    z1=0; z2=Ny-1;
	}
	else
	    if ((z1=atoi(argv[3]))<0 ||
		(z2=atoi(argv[4]))>=Ny || z1>=z2) Error("bad slice range.");

	printf("%d\t%d\n",Nx,Ny);

	N1=Nx-1; N2=Nx/2;

	if (argc==4 || argc==6) {
	    if ((array=(double **)MA(Ny,array))==NULL) Error(cmae);

	    Nc=N2*2+1;
	    for (y=0; y<Ny; y++)
		if ((array[y]=(double *)MA(Nc,*array))==NULL) Error(cmae);
	}

	for (y=0; y<Ny; y++) {
	    RMS0=CalculateRMS(N1,dx0=0,x1=0,x2=Nx,p000[y],p180[y]);
	    if (argc==4 || argc==6) array[y][N2]=RMS0;

	    for (dx=1; dx<=N2; dx++) {
		if ((RMS=CalculateRMS(N1,-dx,dx,Nx,p000[y],p180[y]))<RMS0) {
		    RMS0=RMS; dx0=(-dx); x1=dx; x2=Nx;
		}
		if (argc==4 || argc==6) array[y][N2+dx]=RMS;

		if ((RMS=CalculateRMS(N1,dx,0,Nx-dx,p000[y],p180[y]))<RMS0) {
		    RMS0=RMS; dx0=dx; x1=0; x2=Nx-dx;
		}
		if (argc==4 || argc==6) array[y][N2-dx]=RMS;
	    }
	    R0=(-(double)(N1+dx0)/2.0);

	    CC=0.0; for (x=x1; x<x2; x++) CC+=(p000[y][x+dx0]*p180[y][N1-x]);

	    printf("%d\t%lg\t%lf\t%lf\n",y,R0,RMS0,CC);

	    if (y>=z1 && y<=z2) {
		sCC+=CC;
		sCCy+=(CC*(double)y);
		sCCR0+=(CC*R0);
		sCCy2+=(CC*(double)y*(double)y);
		sCCR02+=(CC*R0*R0);
		sCCyR0+=(CC*(double)y*R0);
	    }
	}
	D=sCCy2*sCC-sCCy*sCCy;
	A=(sCC*sCCyR0-sCCy*sCCR0)/D;
	B=(sCCy2*sCCR0-sCCy*sCCyR0)/D;
	printf("%lf\t%lf\t%lf",A,B,Sqrt((sCCR02-A*sCCyR0-B*sCCR0)/sCC));
	if (argc==5 || argc==6) printf("\t%d\t%d",z1,z2);
	putchar('\n');

	if (argc==4 || argc==6) {
	    RMS1=RMS2=RMS0;
	    for (y=0; y<Ny; y++) for (dx=0; dx<Nc; dx++) {
		if (array[y][dx]<RMS1) RMS1=array[y][dx];
		if (array[y][dx]>RMS2) RMS2=array[y][dx];
	    }
	    dRMS=(RMS2-RMS1)/((double)(1<<BPS)-1.0);

	    cell=(Cell **)array;
	    for (y=0; y<Ny; y++) for (dx=0; dx<Nc; dx++)
		cell[y][dx]=(Cell)floor((array[y][dx]-RMS1)/dRMS+0.5);

	    sprintf(desc,"%d\t%lf\t%lf\n",Nx,RMS1,RMS2);
	    StoreImageFile(argv[argc-1],Nc,Ny,BPS,cell,desc);
	}
	return 0;
}
