
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "rif.h"
#include "sif.h"

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

static int	A2I(str)
char		*str;
{
	double	d=atof(str);

	return (d>=0.0)?(int)floor(d+0.5):(int)ceil(d-0.5);
}

#define DESC_LEN	256

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nx1,Ny1,Nx2,Ny2,dx,dy,BPS,x1,y1,x2,y2,Nx,Ny,x,y;
	Cell	**cell1,**cell2;
	double	v0,v1,v2,*sub,*ptr,v,min,max,base,step;
	char	desc[DESC_LEN];

	if (argc!=11 && argc!=13)
	    Error(
"usage : stvi TIFF1 base1 step1 TIFF2 base2 step2 dx dy {base3 step3} BPS TIFF3"
	    );

	ReadImageFile(argv[1],&Nx1,&Ny1,NULL,&cell1,NULL);
	ReadImageFile(argv[4],&Nx2,&Ny2,NULL,&cell2,NULL);

	v0=atof(argv[5])-atof(argv[2]);
	v1=atof(argv[3]);
	v2=atof(argv[6]);

	if ((dx=A2I(argv[7]))<=(-Nx1) || dx>=Nx2 ||
	    (dy=A2I(argv[8]))<=(-Ny1) || dy>=Ny2) Error("bad distance.");

	if ((BPS=atoi(argv[argc-2]))<=0 || BPS>8*sizeof(Cell))
	    Error("bad bits per sample.");

	if (dx>=0) { x1=0; x2=dx; } else { x1=(-dx); x2=0; }
	if (dy>=0) { y1=0; y2=dy; } else { y1=(-dy); y2=0; }

	Nx=Nx1-x1; x=Nx2-x2; if (Nx>x) Nx=x;
	Ny=Ny1-y1; y=Ny2-y2; if (Ny>y) Ny=y;

	if ((sub=(double *)malloc(Nx*Ny*sizeof(double)))==NULL)
	    Error("memory allocation error.");

	ptr=sub;
	for (y=0; y<Ny; y++)
	for (x=0; x<Nx; x++) {
	    *(ptr++)=v=v0+v2*(double)cell2[y+y2][x+x2]
			 -v1*(double)cell1[y+y1][x+x1];

	    if (x==0 && y==0)
		min=max=v;
	    else
		if (v<min) min=v; else if (v>max) max=v;
	}

	if (argc==11) {
	    base=min;
	    if (min==max) {
		fputs("bad value range (warning).\n",stderr);

		step=1.0;
	    }
	    else
		step=(max-min)/(double)((1<<BPS)-1);

	    sprintf(desc,"%lf\t%lf",min,max);
	}
	else {
	    base=atof(argv[9]);
	    if ((step=atof(argv[10]))==0.0) Error("bad value step.");

	    sprintf(desc,"%lf\t%lf\n%lf\t%lf",min,max,base,step);

	    min=base; max=base+step*(double)((1<<BPS)-1);
	}
	puts(desc);

	ptr=sub;
	for (y=0; y<Ny; y++)
	for (x=0; x<Nx; x++) {
	    if ((v=(*ptr++))<min) v=min; else if (v>max) v=max;

	    cell1[y][x]=(Cell)floor((v-base)/step+0.5);
	}
	StoreImageFile(argv[argc-1],Nx,Ny,BPS,cell1,desc); return 0;
}
