
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "cell.h"
#include "rif.h"
#include "sif.h"

static void	Error(msg)
char		*msg;
{
	(void)fputs(msg,stderr); (void)fputc('\n',stderr); exit(1);
}

#define STR_LEN	256

#define MA(ptr,num)	malloc(sizeof(*ptr)*(num))

static void	ReadNxyzPathBaseStep(Nx,Ny,Nz,path,base,step)
int		*Nx,*Ny,*Nz;
char		***path;
double		**base,**step;
{
	int	z,nx,ny;
	char	str[STR_LEN];

	if (scanf("%d",Nz)!=1 || *Nz<=0) Error("bad slice number.");

	if ((*path=(char **)MA(*path,*Nz))==NULL ||
	    (*base=(double *)MA(*base,*Nz))==NULL ||
	    (*step=(double *)MA(*step,*Nz))==NULL)
	    Error("path, base and/or step memory allocation error.");

	*Nx=(*Ny)=0;
	for (z=0; z<*Nz; z++) {
	    if (scanf("%s %lf %lf",str,*base+z,*step+z)!=3)
		Error("path, base and/or step read error.");

	    ReadImageFile(str,&nx,&ny,NULL,NULL,NULL);
	    if (nx>*Nx) *Nx=nx;
	    if (ny>*Ny) *Ny=ny;

	    if (((*path)[z]=(char *)MA(**path,strlen(str)+1))==NULL)
		Error("path memory allocation error.");

	    (void)strcpy((*path)[z],str);
	}
}

static int	A2I(str)
char		*str;
{
	double	d=atof(str);

	return (d>=0.0)?(int)floor(d+0.5):(int)ceil(d-0.5);
}

static void	ReadNxNyCell(Nx,Ny,path,nx,ny,cell)
int		Nx,Ny,*nx,*ny;
char		*path;
Cell		***cell;
{
	ReadImageFile(path,nx,ny,NULL,cell,NULL);
	if (*nx>Nx || *ny>Ny) Error("too large image.");
}

static void	FreeCell(ny,cell)
int		ny;
Cell		**cell;
{
	while (--ny>=0) free(cell[ny]); free(cell);
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nx1,Ny1,Nz1,Nx2,Ny2,Nz2,Nx,Ny,Nz,dx,dy,dz,BPS,
		x1,y1,z1,x2,y2,z2,x,y,z,nx1,ny1,nx2,ny2,nx,ny;
	char	**path1,**path2,
		slice[STR_LEN],form[STR_LEN],path[STR_LEN],desc[STR_LEN];
	double	*base1,*step1,*base2,*step2,span,base,step,limit,
		*sub,bias,*ptr,val,min,max;
	Cell	**cell1,**cell2;

	if (argc!=6 && argc!=8)
	    Error("usage : stvi3 dx dy dz {base step} BPS directory");

	ReadNxyzPathBaseStep(&Nx1,&Ny1,&Nz1,&path1,&base1,&step1);
	ReadNxyzPathBaseStep(&Nx2,&Ny2,&Nz2,&path2,&base2,&step2);

	Nx=(Nx1<Nx2)?Nx1:Nx2;
	Ny=(Ny1<Ny2)?Ny1:Ny2;
	Nz=(Nz1<Nz2)?Nz1:Nz2;
	if ((dx=A2I(argv[1]))<=(-Nx) || dx>=Nx ||
	    (dy=A2I(argv[2]))<=(-Ny) || dy>=Ny ||
	    (dz=A2I(argv[3]))<=(-Nz) || dz>=Nz) Error("bad distance.");

	if ((BPS=atoi(argv[argc-2]))<=0 || BPS>8*sizeof(Cell))
	    Error("bad bits per sample.");

	span=(double)((1<<BPS)-1);

	if (argc==8) {
	    base=atof(argv[4]);
	    if ((step=atof(argv[5]))==0.0) Error("bad value step.");

	    limit=base+step*span;
	}

	if (dx>=0) { x1=0; x2=dx; } else { x1=(-dx); x2=0; }
	if (dy>=0) { y1=0; y2=dy; } else { y1=(-dy); y2=0; }
	if (dz>=0) { z1=0; z2=dz; } else { z1=(-dz); z2=0; }

	Nx=Nx1-x1; x=Nx2-x2; if (Nx>x) Nx=x;
	Ny=Ny1-y1; y=Ny2-y2; if (Ny>y) Ny=y;
	Nz=Nz1-z1; z=Nz2-z2; if (Nz>z) Nz=z;

	if ((sub=(double *)MA(sub,Nx*Ny))==NULL)
	    Error("data memory allocation error.");

	sprintf(slice,"%d",Nz-1); sprintf(form,"%%0%dd",strlen(slice));

	for (z=0; z<Nz; z++) {
	    sprintf(slice,form,z);

	    ReadNxNyCell(Nx1,Ny1,path1[z+z1],&nx1,&ny1,&cell1);
	    ReadNxNyCell(Nx2,Ny2,path2[z+z2],&nx2,&ny2,&cell2);

	    nx=nx1-x1; x=nx2-x2; if (nx>x) nx=x;
	    ny=ny1-y1; y=ny2-y2; if (ny>y) ny=y;
	    if (nx<=0 || ny<=0) {
		fprintf(stderr,"%s : too small image (skipping).\n",slice);
		continue;
	    }

	    bias=base2[z+z2]-base1[z+z1];

	    ptr=sub;
	    for (y=0; y<ny; y++)
	    for (x=0; x<nx; x++) {
		*(ptr++)=val=bias+step2[z+z2]*(double)cell2[y+y2][x+x2]
				 -step1[z+z1]*(double)cell1[y+y1][x+x1];

		if (x==0 && y==0)
		    min=max=val;
		else
		    if (val<min) min=val; else if (val>max) max=val;
	    }
	    printf("%s\t%lf\t%lf\n",slice,min,max);

	    if (argc==6) {
		base=min;
		if (min==max) {
		    fprintf(stderr,"%s : bad value range (warning).\n",slice);

		    step=1.0;
		}
		else
		    step=(max-min)/span;

		sprintf(desc,"%lf\t%lf",min,max);
	    }
	    else {
		sprintf(desc,"%lf\t%lf\n%lf\t%lf",min,max,base,step);

		min=base; max=limit;
	    }

	    ptr=sub;
	    for (y=0; y<ny; y++)
	    for (x=0; x<nx; x++) {
		if ((val=(*ptr++))<min) val=min; else if (val>max) val=max;

		cell1[y][x]=(Cell)floor((val-base)/step+0.5);
	    }
	    sprintf(path,"%s/%s.tif",argv[argc-1],slice);
	    StoreImageFile(path,nx,ny,BPS,cell1,desc);

	    FreeCell(ny1,cell1); FreeCell(ny2,cell2);
	}
	return 0;
}
