
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "sif.h"

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

typedef struct {
		int	i;
		double	d;
	} Sum;

#define ALLOC(type,size)	(type *)malloc(sizeof(type)*(size))

int	main(argc,argv)
int	argc;
char	**argv;
{
	double	x0,y0,Sx,Sy,X,Y,v,x1,x2,y1,y2,v1,v2,bg,base,step,roof;
	int	Nx,Ny,y,x,BPS,start=1;
	Cell	**cell,max;
	Sum	**sum;
	char	*mae="memory allocation error.";

	if (argc!=10 && argc!=12)
	    Error("usage : xyv2t x0 y0 Sx Sy Nx Ny bg {base step} BPS TIFF");

	x0=atof(argv[1]);
	y0=atof(argv[2]);

	if ((Sx=atof(argv[3]))==0.0 ||
	    (Sy=atof(argv[4]))==0.0) Error("bad pixel size.");

	if ((Nx=atoi(argv[5]))<=0 ||
	    (Ny=atoi(argv[6]))<=0) Error("bad pixel number.");

	if ((cell=(Cell **)(sum=ALLOC(Sum *,Ny)))==NULL) Error(mae);

	for (y=0; y<Ny; y++) {
	    if ((sum[y]=ALLOC(Sum,Nx))==NULL) Error(mae);

	    for (x=0; x<Nx; x++) sum[y][x].d=(double)(sum[y][x].i=0);
	}

	while (scanf("%lf %lf %lf",&X,&Y,&v)==3) {
	    if ((x=(int)floor((X-x0)/Sx+0.5))>=0 && x<Nx &&
		(y=(int)floor((Y-y0)/Sy+0.5))>=0 && y<Ny) {
		++sum[y][x].i; sum[y][x].d+=v;
	    }

	    if (start) {
		start=0; x1=x2=X; y1=y2=Y; v1=v2=v;
	    }
	    else {
		if (X<x1) x1=X; else if (X>x2) x2=X;
		if (Y<y1) y1=Y; else if (Y>y2) y2=Y;
		if (v<v1) v1=v; else if (v>v2) v2=v;
	    }
	}
	if (start) Error("no data.");

	printf("%lf\t%lf\n%lf\t%lf\n%lf\t%lf\n",x1,x2,y1,y2,v1,v2);

	if ((BPS=atoi(argv[argc-2]))<1 || BPS>8*sizeof(Cell)) Error("bad BPS.");

	max=(Cell)((1<<BPS)-1);

	if (argc==10) {
	    if (v1==v2) fputs("no value range (warning).\n",stderr);

	    step=(v2-v1)/(double)max;
	    base=v1;
	    roof=v2;
	}
	else {
	    if ((step=atof(argv[9]))<=0.0) Error("bad step.");

	    base=atof(argv[8]);
	    roof=base+step*(double)max;
	}

	bg=atof(argv[7]);

	for (y=0; y<Ny; y++)
	for (x=0; x<Nx; x++) {
	    v=(sum[y][x].i)?sum[y][x].d/(double)sum[y][x].i:bg;
	    cell[y][x]=(v<base)?0:(v>roof)?max:(Cell)floor((v-base)/step+0.5);
	}

	StoreImageFile(argv[argc-1],Nx,Ny,BPS,cell,NULL);

	return 0;
}
