
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "sif.h"

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

#define STR_LEN	256

typedef struct {
		int	i;
		double	d;
	} Sum;

#define ALLOC(type,size)	(type *)malloc(sizeof(type)*(size))

int	main(argc,argv)
int	argc;
char	**argv;
{
	double	x0,y0,z0,Sx,Sy,Sz,X,Y,Z,v,x1,x2,y1,y2,z1,z2,v1,v2,
		bg,base,step,roof;
	int	Nx,Ny,Nz,z,y,x,BPS,len,start=1;
	Cell	***cell,max;
	Sum	***sum;
	char	form[STR_LEN],path[STR_LEN],
		*mae="memory allocation error.";

	if (argc!=13 && argc!=15)
Error("usage : xyzv2t x0 y0 z0 Sx Sy Sz Nx Ny Nz bg {base step} BPS directory");

	x0=atof(argv[1]);
	y0=atof(argv[2]);
	z0=atof(argv[3]);

	if ((Sx=atof(argv[4]))==0.0 ||
	    (Sy=atof(argv[5]))==0.0 ||
	    (Sz=atof(argv[6]))==0.0) Error("bad pixel size.");

	if ((Nx=atoi(argv[7]))<=0 ||
	    (Ny=atoi(argv[8]))<=0 ||
	    (Nz=atoi(argv[9]))<=0) Error("bad pixel number.");

	if ((cell=(Cell ***)(sum=ALLOC(Sum **,Nz)))==NULL) Error(mae);

	for (z=0; z<Nz; z++) {
	    if ((sum[z]=ALLOC(Sum *,Ny))==NULL) Error(mae);

	    for (y=0; y<Ny; y++) {
		if ((sum[z][y]=ALLOC(Sum,Nx))==NULL) Error(mae);

		for (x=0; x<Nx; x++) sum[z][y][x].d=(double)(sum[z][y][x].i=0);
	    }
	}

	while (scanf("%lf %lf %lf %lf",&X,&Y,&Z,&v)==4) {
	    if ((x=(int)floor((X-x0)/Sx+0.5))>=0 && x<Nx &&
		(y=(int)floor((Y-y0)/Sy+0.5))>=0 && y<Ny &&
		(z=(int)floor((Z-z0)/Sz+0.5))>=0 && z<Nz) {
		++sum[z][y][x].i; sum[z][y][x].d+=v;
	    }

	    if (start) {
		start=0; x1=x2=X; y1=y2=Y; z1=z2=Z; v1=v2=v;
	    }
	    else {
		if (X<x1) x1=X; else if (X>x2) x2=X;
		if (Y<y1) y1=Y; else if (Y>y2) y2=Y;
		if (Z<z1) z1=Z; else if (Z>z2) z2=Z;
		if (v<v1) v1=v; else if (v>v2) v2=v;
	    }
	}
	if (start) Error("no data.");

printf("%lf\t%lf\n%lf\t%lf\n%lf\t%lf\n%lf\t%lf\n",x1,x2,y1,y2,z1,z2,v1,v2);

	if ((BPS=atoi(argv[argc-2]))<1 || BPS>8*sizeof(Cell)) Error("bad BPS.");

	max=(Cell)((1<<BPS)-1);

	if (argc==13) {
	    if (v1==v2) fputs("no value range (warning).\n",stderr);

	    step=(v2-v1)/(double)max;
	    base=v1;
	    roof=v2;
	}
	else {
	    if ((step=atof(argv[12]))<=0.0) Error("bad step.");

	    base=atof(argv[11]);
	    roof=base+step*(double)max;
	}

	bg=atof(argv[10]);

	len=0; for (z=Nz-1; z>0; z/=10) len++;
	sprintf(form,"%s/%%0%dd.tif",argv[argc-1],(len)?len:1);

	for (z=0; z<Nz; z++) {
	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++) {
		v=(sum[z][y][x].i)?sum[z][y][x].d/(double)sum[z][y][x].i:bg;
	cell[z][y][x]=(v<base)?0:(v>roof)?max:(Cell)floor((v-base)/step+0.5);
	    }

	    sprintf(path,form,z);
	    StoreImageFile(path,Nx,Ny,BPS,cell[z],NULL);
	}

	return 0;
}
